# not bad --> almost there --> see what is really missing and add it for the MS --> TODO


# TODO I have a big bug in the polarity --> need fix it but maybe not very urgent in fact !!!
# do a small test on a small image --> easy to debug
# test of all

# examples of SQLite commands I'd like to reimplement rapidly
#           PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, area_cells FROM cells + LUT DNA
#             PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, S0_stretch_cells FROM cells WHERE is_border_cell=='false' + LUT DNA + OPACITY 35% + DILATATION 1
#             PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, S0_stretch_cells FROM cells NATURAL JOIN tracked_clone_022 + LUT DNA MIN=GLOBAL_MIN MAX=GLOBAL_MAX + OPACITY 35% + DILATATION 1
#             PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells FROM cells ORDER BY area_cells DESC LIMIT 10 + LUT DNA MIN=GLOBAL_MIN MAX=GLOBAL_MAX
#             PLOT AS ARROWS SELECT first_pixel_x_cells,first_pixel_y_cells , center_x_cells,center_y_cells, area_cells AS COLOR, '1' AS STROKE_SIZE, 'DEFAULT_ARROW' AS ARROWHEAD_TYPE, '5' AS ARROWHEAD_HEIGHT,'3' AS ARROWHEAD_HEIGHT, '0.3' AS STROKE_OPACITY FROM cells + LUT DNA
#             PLOT AS BONDS SELECT first_pixel_x_bonds, first_pixel_y_bonds, bond_orientation FROM bonds WHERE first_pixel_y_bonds != 'null' + LUT DNA


# some columns would require specific plots
# implement plot as nematic
# implement plot as cells


# offer  exclude border cells
# support quantiles
#


# https://matplotlib.org/2.0.2/faq/usage_faq.html --> check backends beacuse that is most likely what I need for my figure editor


# plots stuff as cell, bonds or vertices ideally from a db
# should not be too hard --> need load the image duplicate it and do the plots through conversions maybe need enable numba otherwise it will be too slow!!!
# from PIL.ImageDraw import ImageDraw
import traceback

from PIL import ImageDraw, Image

from epyseg.draw.shapes.image2d import Image2D
from epyseg.img import Img, PIL_to_numpy
from skimage.measure import regionprops
import numpy as np
import matplotlib
from scipy.ndimage import grey_erosion, grey_dilation, generate_binary_structure
import os
from epyseg.ta.database.sql import TAsql, createMasterDB, sort_col_numpy
from epyseg.ta.luts.lut_minimal_test import PaletteCreator, apply_lut
from epyseg.ta.measurements.nematic import Nematic
from epyseg.ta.tracking.tools import smart_name_parser


# matplotlib.use('Qt5Agg') #, Qt5Cairo,
# matplotlib.use('Qt5Cairo') #, Qt5Cairo,
# matplotlib.use('WXAgg') #, Qt5Cairo,
# matplotlib.use('agg') #, Qt5Cairo,
# matplotlib.use('TkCairo') #, Qt5Cairo,
#  supported values are['GTK3Agg', 'GTK3Cairo', 'MacOSX', 'nbAgg', 'Qt4Agg', 'Qt4Cairo', 'Qt5Agg', 'Qt5Cairo', 'TkAgg', 'TkCairo', 'WebAgg', 'WX', 'WXAgg', 'WXCairo', 'agg', 'cairo', 'pdf', 'pgf', 'ps', 'svg', 'template']
# matplotlib.rcParams['backend'] = 'cairo'

# - interactive
# backends:
# GTK3Agg, GTK3Cairo, MacOSX, nbAgg,
# Qt4Agg, Qt4Cairo, Qt5Agg, Qt5Cairo,
# TkAgg, TkCairo, WebAgg, WX, WXAgg, WXCairo
#
# - non - interactive
# backends:
# agg, cairo, pdf, pgf, ps, svg, template

# print(matplotlib.use)

import matplotlib.pyplot as plt
from epyseg.tools.logger import TA_logger  # logging
from epyseg.utils.loadlist import  loadlist

logger = TA_logger()

# TODO finalize that
# need a type of plot and need data to be plotted that is associated to the id of the relevant table --> TODO
# whole data can also be obtained from a SQL command --> such as PLOT AS CELLS ...
# can I do blending directly at the qimage level ??? if so --> quite simple in a way!!!


# then need get a mapping of the colors --> TODO
# for the SQL plots see how I can do that --> need convert to number and do stuff with it
# faire un importeur de csv
# plus simple de matcher la cell ID avec une valeur que les coordonnées en fait dans cette nouvelle version et en plus il y a moins de données ce qui sera plus simple à gérer -> juste besoin de la local ID des cellules ou des vertex ou des bonds --> vraiment du gateau


'''
TODO to be ok with TA

//PLOT AS BONDS SQL
//PLOT AS RECTANGLE
//PLOT AS LINE
//PLOT AS NEMATIC
//PLOT AS VECTOR same as plot as line
//PLOT AS SQUARE
//PLOT AS POINTS
PLOT AS STRINGS 

        plotCommandsTranslator.put("[Cc][Ee][Ll][Ll][Ss]{0,1}", "CELLS");
        plotCommandsTranslator.put("[Bb][Oo][Nn][Dd][Ss]{0,1}", "BONDS");
        plotCommandsTranslator.put("[Vv][Ee][Cc][Tt][Oo][Rr][Ss]{0,1}", "VECTORS");//vectors
        plotCommandsTranslator.put("[Nn][Ee][Mm][Aa][Tt][Ii][Cc][Ss]{0,1}", "NEMATICS");
        plotCommandsTranslator.put("[Ss][Qq][Uu][Aa][Rr][Ee][Ss]{0,1}", "SQUARES");
        plotCommandsTranslator.put("[Rr][Ee][Cc][Tt][Aa][Nn][Gg][Ll][Ee][Ss]{0,1}", "RECTANGLES");
        plotCommandsTranslator.put("[Ee][Ll][Ll][Ii][Pp][Ss][Ee][Ss]{0,1}", "ELLIPSES");
        plotCommandsTranslator.put("[Cc][Ii][Rr][Cc][Ll][Ee][Ss]{0,1}", "CIRCLES");//should we cast it as an ellipse
        plotCommandsTranslator.put("[Pp][Oo][Ll][Yy][Gg][Oo][Nn][Ss]{0,1}", "POLYGONS");
        plotCommandsTranslator.put("[Ll][Ii][Nn][Ee][Ss]{0,1}", "LINES");
        plotCommandsTranslator.put("[Aa][Rr][Rr][Oo][Ww][Ss]{0,1}", "ARROWS");//default arrow no rescaling allowed for this unlike for the nematics
        plotCommandsTranslator.put("[Pp][Oo][Ii][Nn][Tt][Ss]{0,1}|[Dd][Oo][Tt][Ss]{0,1}", "DOTS");//points or DOTS
        plotCommandsTranslator.put("[Ss][Tt][Rr][Ii][Nn][Gg][Ss]{0,1}|[Tt][Ee][Xx][Tt][Ss]{0,1}", "STRINGS");//points or DOTS


    need implement this too
         if (commandTypeAndCommand.containsKey("EROSION")) {
            parsedPlot[4] = "EROSION " + commandTypeAndCommand.get("EROSION");
        }
        if (commandTypeAndCommand.containsKey("DILATATION")) {
//            System.out.println("in " +commandTypeAndCommand.get("DILATATION"));
            parsedPlot[4] = "DILATATION " + commandTypeAndCommand.get("DILATATION");
        }
        //now we further parse lut flavours to do things with them
        if (commandTypeAndCommand.containsKey("LUT")) {
            parsedPlot[3] = "LUT " + commandTypeAndCommand.get("LUT");
        }
        if (commandTypeAndCommand.containsKey("OPACITY")) {
            parsedPlot[2] = "OPACITY " + commandTypeAndCommand.get("OPACITY");
        }
        if (commandTypeAndCommand.containsKey("TRANSPARENCY")) {
            parsedPlot[2] = "TRANSPARENCY " + commandTypeAndCommand.get("TRANSPARENCY");
        }
        
        
        # see how I can parse it smartly --> now
        
        
        
        
        some tests that need be implemented
        
           String testAdvancedCMD = "PLOT AS CELLS, area_cells+10 SELECT * FROM cells + OPACITY 55% + LUT (255*R,133+G,23+B)";

            //sinon juste faire des parsers de LUTS
            //styles LUT name ou formule des correspondances entre 
            //si formule --> entre parentheses
            HashMap<String, String> LUT_ParserNType = new HashMap<String, String>();

            String testPercentages = "TEST 55% 63% #FF00FF 33%";
            System.out.println(BrickPlotter.getPercent(testPercentages));

            String colorSplitterTest = "TEST #00FF00 #FF0000 #0000AA";
            System.out.println(BrickPlotter.parseHTMLColors(colorSplitterTest));

            System.out.println(BrickPlotter.parseHTMLColors("LUT #FF00FF 0:0 255:16000 12:#FF0000 16:1500"));



# manually force max and min too --> not sure it is very necessary 

 String maxParser = ".*([Mm][Aa][Xx]\\s{0,}[=]\\s{0,}(\\d{1,}[\\.,]{0,}\\d{0,})).*";//
            String minParser = ".*([Mm][Ii][Nn]\\s{0,}[=]\\s{0,}(\\d{1,}[\\.,]{0,}\\d{0,})).*";//
            String testCommand = "+ LUT DNA max=123.3 min=10.1";


    and I should also be able to parse such a complex line --> DOES A COMBINATION OF ALL
   String command = "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, area_cells FROM cells + LUT DNA +OPACITY 35% #FF0000\n"
                + "PLOT AS CELLS   SELECT * FROM cells \n"
                + "#SQL=SELECT * FROM bonds\n"
                + "CREATE TABLE TEST2 AS SELECT * FROM cells\n"
                + "#SQL=SELECT * FROM bonds\n"//ignore commented lines
                + "#SQL =  SELECT * FROM bonds\n"
                + "  PLOT   AS      BONDS  SELECT * FROM SELECT * FROM bonds   \n"
                + "  PLOT   AS      CRABOUNIASSE  SELECT * FROM cells  \n"
                + "PLOT AS  LINES SELECT * FROM cells\n"
                + "PLOT AS VECTORS   SELECT * FROM polarity    "
                + "PLOT AS VECTORS   SELECT * FROM polarity  + OPACITY 35% + LUT RAINBOW \n"
                + "PLOT AS VECTORS SELECT first_pixel_x_cells,first_pixel_y_cells , center_x_cells,center_y_cells, '16166' AS COLORS, '1' AS STROKE_SIZE, 'DEFAULT_ARROW' AS ARROWHEAD_TYPE, '5' AS ARROWHEAD_HEIGHT,'3' AS ARROWHEAD_HEIGHT, '0.3' AS STROKE_OPACITY FROM cells +DILATATION -10\n"
                + "PLOT AS VECTORS SELECT first_pixel_x_cells,first_pixel_y_cells , center_x_cells,center_y_cells, '16166' AS COLORS, '1' AS STROKE_SIZE, 'DEFAULT_ARROW' AS ARROWHEAD_TYPE, '5' AS ARROWHEAD_HEIGHT,'3' AS ARROWHEAD_HEIGHT, '0.3' AS STROKE_OPACITY FROM cells + OPACITY 100%\n"
                + "PLOT AS VECTORS SELECT first_pixel_x_cells,first_pixel_y_cells , center_x_cells,center_y_cells, area_cells AS COLOR, '1' AS STROKE_SIZE, 'DEFAULT_ARROW' AS ARROWHEAD_TYPE, '5' AS ARROWHEAD_HEIGHT,'3' AS ARROWHEAD_HEIGHT, '0.3' AS STROKE_OPACITY FROM cells + LUT DNA\n";
        
        
        
        
        HashMap<String, String> extraParser = new HashMap<String, String>();
        extraParser.put(LUT_flavour, "LUT");
        extraParser.put(OPACITY_flavour, "OPACITY");
        extraParser.put(TRANSPARENCY_flavour, "TRANSPARENCY");
        extraParser.put(DILATATION_flavour, "DILATATION");        

   public static ArrayList<Float> getPercent(String textToParse) {
        ArrayList<Float> percents = new ArrayList<Float>();
        String percentParser = "(\\d{1,})%";
        Pattern p = Pattern.compile(percentParser, Pattern.DOTALL);
        Matcher m = p.matcher(textToParse);
        while (m.find()) {
            percents.add(CommonClasses.String2Float(m.group(1)) / 100f);
        }
        return percents;
    }


public static ArrayList<Integer> parseHTMLColors(String textToParse) {
        /**
         * we parse colors except if preceded by a : stuff
         */
        String colorTAGS = "[^:](#[0-9a-fA-F]{6}|[0-9\\.]{1,})";
        Pattern p = Pattern.compile(colorTAGS, Pattern.DOTALL);
        Matcher m = p.matcher(textToParse);

        ArrayList<Integer> forbiddenColors = new ArrayList<Integer>();
        while (m.find()) {
            //System.out.println(m.group(1));
            forbiddenColors.add(CommonClasses.getColorFromHtmlColor(m.group(1)));
        }
        return forbiddenColors;
    }

   /**
     * deep parsing of LUT text files can do everything with that
     *
     * @param testLUT2
     * @return
     */
    public static HashMap<Integer, Integer> parseLUT(String testLUT2) {
        HashMap<Integer, Integer> LUTconversionTable = new HashMap<Integer, Integer>();
        String LutHumanSpecified = "(\\d{1,}):[']{0,1}([#A-Fa-f0-9]{1,}|[0-9\\.]{1,})[']{0,1}";

        Pattern p = Pattern.compile(LutHumanSpecified);
        Matcher m = p.matcher(testLUT2);

        while (m.find()) {
            try {
                String outputValue = m.group(2);
                if (!outputValue.contains("#")) {
                    LUTconversionTable.put(CommonClasses.String2Int(m.group(1)), CommonClasses.String2Int(outputValue));
                } else {
                    LUTconversionTable.put(CommonClasses.String2Int(m.group(1)), CommonClasses.getColorFromHtmlColor(outputValue));
                }
            } catch (Exception e) {
            }
        }
        return LUTconversionTable;
    }

    private String getSupportedPlotTypes() {
        String supportedPlots = "Supported plots are:";
        for (Map.Entry<String, String> entry : plotCommandsTranslator.entrySet()) {
            String value = entry.getValue();
            supportedPlots += "\n" + value;
        }
        return supportedPlots + "\n";
    }

    public String[] parsedSinglePlotLine(String individualCommmand) {
        if (individualCommmand == null || individualCommmand.isEmpty()) {
            return null;
        }
        /**
         * first we clean the commands
         */
        individualCommmand = individualCommmand.trim();

        /**
         * we ignore all commented lines
         */
        if (individualCommmand.startsWith("#") || individualCommmand.startsWith("//")) {
            return null;
        }
        /**
         * one is plot, two is SQL command, 3 is transparency, four is LUT
         */
        String[] parsedPlot = new String[5];
//        System.out.println(individualCommmand);
//        if (individualCommmand.matches(SQLPattern)) {
//            //parse SQL command
//            String SQLcommand = individualCommmand.replaceAll(SQLPattern, "$1").trim();
////            System.out.println("extracted SQLcommand='" + SQLcommand + "'");
//        }

        HashMap<String, String> commandTypeAndCommand = new HashMap<String, String>();
        //TODO do not cut if in between brackets --> allows formulas to be passed
        //only cuts if + is not in between brackets

        /**
         * first we strip TRANSPARENCY and anaestetic commands
         */
        String LUT_flavour = ".*\\s{0,}[Ll][Uu][Tt][Ss]{0,1}(.*)|.*\\s{0,}[Pp][Aa][Ll][Ee][Tt][Tt][Ee][Ss]{0,1}(.*)";
        String TRANSPARENCY_flavour = ".*\\s{0,}[Tt][Rr][Aa][Nn][Ss][Pp][Aa][Rr][Ee][Nn][Cc][Yy](.*\\d{1,}%.*)";
        String OPACITY_flavour = ".*\\s{0,}[Oo][Pp][Aa][Cc][Ii][Tt][Yy](.*\\d{1,}%.*)";
        String DILATATION_flavour = ".*\\s{0,}[Dd][Ii][Ll][Aa][Tt][Aa][Tt][Ii][Oo][Nn][Ss]{0,1}(.*)|.*\\s{0,}[Ee][Rr][Oo][Ss][Ii][Oo][Nn][Ss]{0,1}(.*)";
        //pour les LUTs faire aussi un parser manuel
        HashMap<String, String> extraParser = new HashMap<String, String>();
        extraParser.put(LUT_flavour, "LUT");
        extraParser.put(OPACITY_flavour, "OPACITY");
        extraParser.put(TRANSPARENCY_flavour, "TRANSPARENCY");
        extraParser.put(DILATATION_flavour, "DILATATION");

        String[] ExtraSplitted = individualCommmand.split("\\+(?![^()]*\\))", -1);
        //System.out.println(ExtraSplitted.length + " "+individualCommmand);
        for (String string : ExtraSplitted) {
            if (string.equals(individualCommmand)) {
                break;
            }
            for (Map.Entry<String, String> entry : extraParser.entrySet()) {
                String key = entry.getKey();
                String value = entry.getValue();
                //System.out.println(key + " "+string + " "+string.matches(key));

                if (string.matches(key)) {
                    commandTypeAndCommand.put(value, string.replaceAll(key, "$1").trim());
                    //            System.out.println(string);
                    individualCommmand = individualCommmand.replaceAll("\\+" + string, "").trim();
                }
            }
        }

        if (commandTypeAndCommand.containsKey("EROSION")) {
            parsedPlot[4] = "EROSION " + commandTypeAndCommand.get("EROSION");
        }
        if (commandTypeAndCommand.containsKey("DILATATION")) {
//            System.out.println("in " +commandTypeAndCommand.get("DILATATION"));
            parsedPlot[4] = "DILATATION " + commandTypeAndCommand.get("DILATATION");
        }
        //now we further parse lut flavours to do things with them
        if (commandTypeAndCommand.containsKey("LUT")) {
            parsedPlot[3] = "LUT " + commandTypeAndCommand.get("LUT");
        }
        if (commandTypeAndCommand.containsKey("OPACITY")) {
            parsedPlot[2] = "OPACITY " + commandTypeAndCommand.get("OPACITY");
        }
        if (commandTypeAndCommand.containsKey("TRANSPARENCY")) {
            parsedPlot[2] = "TRANSPARENCY " + commandTypeAndCommand.get("TRANSPARENCY");
        }

        //TODO further reparse them if needed
        // System.out.println("indi"+ individualCommmand);
        // System.out.println(commndTypeAndCommand);
        if (individualCommmand.matches(plotCommandPattern)) {
//            System.out.println("parsed command");
            //parse SQL command
            String plotType = individualCommmand.replaceAll(plotCommandPattern, "$1").trim();
            /**
             * we check if the plot type matches one of the possible plot types
             * I know
             */
            boolean matched = false;
            for (Map.Entry<String, String> entry : plotCommandsTranslator.entrySet()) {
                String key = entry.getKey();
                if (plotType.matches(key)) {
                    plotType = entry.getValue();
                    matched = true;
                    break;
                }
            }
            if (!matched) {
                System.err.println("UNKNOWN PLOT TYPE: '" + plotType + "'");
                System.err.println(getSupportedPlotTypes());
                plotType = "UNKNOWN";
            }
//            System.out.println("plot type='" + plotType + "'");
            String SQLcommand = individualCommmand.replaceAll(plotCommandPattern, "$2").trim();
//            System.out.println("SQLcommand='" + SQLcommand + "'");

            parsedPlot[0] = plotType;
            parsedPlot[1] = SQLcommand;
//                SQLCommands.add(SQLcommand);
//                plots.add(plotType);
            //try plottings 
        } else {
            /**
             * execute the unmatched pattern as an SQL command
             */
            parsedPlot[1] = individualCommmand;
        }
        return parsedPlot;
        //now we perform the extraction
    }

'''

# check a set of TA commands to see how I can parse them
# in a way the SQL command is anything between select and the end of the line or a plus --> it is anything before the first encountered extra keyword --> should not be too hard to do!!!
# especially now that the command is cleaned
# if unknown keyword --> then return an error


extra_plot_commands_for_images = ['LUT', 'EROSION', 'DILATATION', 'DILATION', 'OPACITY', 'TRANSPARENCY']


def plot_SQL(SQL_command):
    # just get the word after
    # or split and get all words --> very easy TODO in fact
    # if SQL_command.lower().starts_with()

    if SQL_command is None:
        logger.error('Empty SQL command --> nothing to do')
        return

    if not isinstance(SQL_command, str):
        logger.error('SQL command must be a string, instead a ' + str(type(SQL_command)) + ' was passed...')
        return

    if SQL_command.strip().startswith('#') or SQL_command.strip().startswith('//'):
        # assume this is a comment and then nothing TODO
        return

    cleaned_SQL_command = _clean_string(SQL_command)

    print('cleaned_SQL_command "' + cleaned_SQL_command + '"')

    plot_type = None
    # First we get the type of plot then we should get the SQL command
    if cleaned_SQL_command.upper().startswith('PLOT AS CELLS'):
        # plot as cells --> need map data to the local ID database of cells
        # print('plot as cells')
        plot_type = 'CELLS'
        # cleaned_SQL_command.replace()
        # get index and cut after

        # cleaned_SQL_command = cleaned_SQL_command.upper().index('PLOT AS CELLS')
        cleaned_SQL_command = _strip_from_command(cleaned_SQL_command, 'PLOT AS CELLS')
    elif cleaned_SQL_command.upper().startswith('PLOT AS BONDS'):
        # print('plot as bonds')
        plot_type = 'BONDS'
        cleaned_SQL_command = _strip_from_command(cleaned_SQL_command, 'PLOT AS BONDS')
    elif cleaned_SQL_command.upper().startswith('PLOT AS VERTICES'):
        # print('plot as vertices')
        plot_type = 'VERTICES'
        cleaned_SQL_command = _strip_from_command(cleaned_SQL_command, 'PLOT AS VERTICES')
    else:
        # unknown/unsupported plot --> raise an error
        logger.error('unsupported plot command: ' + str(cleaned_SQL_command))
        return

    print('plot_type "' + str(plot_type) + '"')
    # print('cleaned_SQL_command',cleaned_SQL_command)

    # now we need to get the pure SQL command --> see how I do that in TA
    # further parse it --> get the command between different stuff
    # SQL_command is actually anything before the first magic keyword encountered if any --> should be easy TODO

    # get extras

    first_idx = None
    for extra in extra_plot_commands_for_images:
        try:
            idx = cleaned_SQL_command.upper().index(extra)
            # if idx !=-1:
            if first_idx is None:
                first_idx = idx
            else:
                first_idx = min(first_idx, idx)
        except:
            # extra command not found --> ignoring
            continue

    # print('first_idx',first_idx)

    extra_commands_to_parse = None
    if first_idx is not None:
        # cut after and before --> before is the SQL command otherwise ignore
        # very good in fact

        SQL_command = cleaned_SQL_command[0:first_idx].strip()
        if SQL_command.endswith('+'):
            SQL_command = SQL_command[:-1].strip()
        extra_commands_to_parse = cleaned_SQL_command[first_idx:].strip()


    else:
        SQL_command = cleaned_SQL_command

    print('SQL_command "' + str(SQL_command) + '"')
    print('extra_commands_to_parse "' + str(
        extra_commands_to_parse) + '"')  # extras are indicated by a plus --> this way it is very simple to get them --> may remove plus if ends with it

    if extra_commands_to_parse:
        # need further parse these commands to get what I really need --> first find the identifier then find the values
        # print('TODO --> need parse extras')

        # we clean the extras and then get their values
        # extra_commands_to_parse = extra_commands_to_parse.replace('+','')# remove the pluses from extras

        if '+' in extra_commands_to_parse:
            individual_extras = extra_commands_to_parse.split('+')
        else:
            individual_extras = [extra_commands_to_parse]

        for individual_extra in individual_extras:
            print('individual_extra "' + individual_extra.strip() + '"')

            # ça marche super --> now I need to further parse the extras but I'm almost there in fact --> that is very easy TODO

            # split it to get the first keyword then see later how to handle the rest
            extra_type = individual_extra.strip().split()

            # print("extra_type",extra_type)

            if extra_type[0].strip().upper() not in extra_plot_commands_for_images:
                logger.error('unknown image extra ' + str(extra_type[0]) + ' --> ignoring')
            else:
                # known extra --> need parse it to make it meaningful --> TODO
                # need Call a different function depending on the data --> TODO
                if extra_type[0].strip().upper() == extra_plot_commands_for_images[0]:
                    # ['LUT', 'EROSION', 'DILATATION','DILATION', 'OPACITY', 'TRANSPARENCY']
                    print('LUT to parse')
                    print('param', parse_lut(extra_type))
                elif extra_type[0].strip().upper() == extra_plot_commands_for_images[1]:
                    print('EROSION to parse')
                    print('param', parse_erosion_dilation(extra_type))
                elif extra_type[0].strip().upper() == extra_plot_commands_for_images[2] or extra_type[
                    0].strip().upper() == extra_plot_commands_for_images[3]:
                    print('DILATION to parse')
                    print('param', parse_erosion_dilation(extra_type))
                elif extra_type[0].strip().upper() == extra_plot_commands_for_images[4]:
                    print('OPACITY to parse')
                    print('param', parse_opacity(extra_type))
                elif extra_type[0].strip().upper() == extra_plot_commands_for_images[5]:
                    print('TRANSPARENCY to parse')
                    print('param', 1. - parse_opacity(extra_type))

                # else:
                #     print('weird extra', extra_type[0])

        # --> very good idea

        # extra_commands_to_parse = _clean_string(extra_commands_to_parse)
        # print('cleaned_extras',extra_commands_to_parse)
        # now need parse it --> can do it in a consecutive way
        # if unknwon keyword --> say it and move on to next step
        # should not be too hard

        # then need even further parse it to get the real value of the parameter
        # extra_settings = extra_commands_to_parse.split()
        # if keyword is known and until next keyword --> get the values --> pb how can I warn about unknown keywords!!!

        # for extra_setting in extra_settings:
        #     print(extra_setting)


'''
    public static HashMap<Integer, Integer> parseLUT(String testLUT2) {
        HashMap<Integer, Integer> LUTconversionTable = new HashMap<Integer, Integer>();
        String LutHumanSpecified = "(\\d{1,}):[']{0,1}([#A-Fa-f0-9]{1,}|[0-9\\.]{1,})[']{0,1}";

        Pattern p = Pattern.compile(LutHumanSpecified);
        Matcher m = p.matcher(testLUT2);

        while (m.find()) {
            try {
                String outputValue = m.group(2);
                if (!outputValue.contains("#")) {
                    LUTconversionTable.put(CommonClasses.String2Int(m.group(1)), CommonClasses.String2Int(outputValue));
                } else {
                    LUTconversionTable.put(CommonClasses.String2Int(m.group(1)), CommonClasses.getColorFromHtmlColor(outputValue));
                }
            } catch (Exception e) {
            }
        }
        return LUTconversionTable;
    }
'''


def parse_lut(extra_type):
    # ['LUT', 'DNA', 'MIN=GLOBAL_MIN', 'MAX=GLOBAL_MAX'] # do I need to reimplement that ??? not sure it is very ueseful as the SQL command can easily be edited to avoid that
    # print(extra_type)
    # print(len(extra_type), not ':' in extra_type[1])
    if len(extra_type) == 2 and (not ':' in extra_type[1]):
        try:
            return extra_type[1]  # returns the name of the LUT
        except:
            pass
    else:
        if ':' in extra_type[1]:
            # get lutvalues and return them as a dict
            lut_mapping = {}
            for mapping in extra_type[1:]:
                split_mappings = mapping.split(':')
                split_mappings = [split_mapping.strip() for split_mapping in split_mappings]
                key = int(split_mappings[0])
                value = split_mappings[1]
                if '#' in value:
                    # convert it back to integer
                    value = int(value.replace('#', ''),
                                16)  # convert hexadecimal (base 16) back to int --> must remove #
                else:
                    value = int(value)
                lut_mapping[key] = value
        else:
            # need handle a stuff with max min or something hybrid in fact
            # TODO implement parsing of max and min but ignore for now ['LUT', 'DNA', 'MIN=GLOBAL_MIN', 'MAX=GLOBAL_MAX']
            logger.warning('LUT bounds not implemented yet')
            return parse_lut(extra_type=extra_type[0:2])
        return lut_mapping
    # TODO
    # something went wrong --> ignoring
    return None


def parse_opacity(extra_type):
    # TODO
    if len(extra_type) == 2:
        try:
            ispercent = '%' in extra_type[1]
            if ispercent:
                extra_type[1] = extra_type[1].replace('%', '')
            value = float(extra_type[1])
            if ispercent:
                value /= 100.
            return value  # returns the name of the LUT
        except:
            pass
    return None


def parse_erosion_dilation(extra_type):
    # TODO
    if len(extra_type) == 2:
        try:
            return int(extra_type[1])  # returns the name of the LUT
        except:
            pass
    return None


def _clean_string(str):
    parsed_SQL_commands = str.split()
    parsed_SQL_commands = [parsed_SQL_command.strip() for parsed_SQL_command in parsed_SQL_commands]
    # recreate it as a single sentence and do the tests
    cleaned_SQL_command = ' '.join(parsed_SQL_commands)
    return cleaned_SQL_command


def _strip_from_command(command, things_to_strip):
    # begin_strip = command.upper().index(things_to_strip)
    # print('begin_strip', begin_strip)
    # end_strip =len(things_to_strip)
    # print('end_strip',end_strip)

    # return command/
    # command = command[begin_strip:end_strip]
    # print('stripped "' +command+'"')

    # print(text.removesuffix('ly'))
    # print(text.removesuffix('World'))

    import re
    case_insensitive_replace = re.compile(re.escape(things_to_strip), re.IGNORECASE)
    stripped = case_insensitive_replace.sub('', command).strip()

    # print(stripped)
    return stripped

# **kwargs are the extra params --> TODO also ask for opacity
# indeed bg image must be provided there maybe also with a channel --> TODO--> see how I can handle that?

# make it store the max and the minb


# NB THERE IS A BUG THAT FORCES THE PLOT TWICE --> PROBABLY SOME ADJUSTMENT STUFF --> NEED CHANGE THIS
def plot_as_any(parent_image, SQL_command, plot_type='cells', return_mask=False, invert_mask=True, db=None, current_frame=None, **kwargs):
    if parent_image is None:
        logger.error('No input image specified, cannot plot image!')
        return

    # print('current_frame inside', current_frame)

    TA_path = smart_name_parser(parent_image, ordered_output='TA')
    erosion_dilation = None
    lut = None
    opacity = None
    freq = None

    if kwargs:
        # parse extra parameters such as dilation...
        # ['LUT', 'EROSION', 'DILATATION','DILATION', 'OPACITY', 'TRANSPARENCY']
        if extra_plot_commands_for_images[0] in kwargs:
            lut = kwargs[extra_plot_commands_for_images[0]]
        if extra_plot_commands_for_images[1] in kwargs:
            erosion_dilation = -kwargs[extra_plot_commands_for_images[1]]
        if extra_plot_commands_for_images[
            3] in kwargs:  # elif extra_type[0].strip().upper() == extra_plot_commands_for_images[2]
            erosion_dilation = kwargs[extra_plot_commands_for_images[3]]
        if extra_plot_commands_for_images[4] in kwargs:
            opacity = kwargs[extra_plot_commands_for_images[4]]
        if extra_plot_commands_for_images[5] in kwargs:
            opacity = 1. - kwargs[extra_plot_commands_for_images[5]]
        if 'freq' in kwargs:
            freq = kwargs['freq']

    # those guys require the cell db
    if plot_type in ['cells', 'nematic', 'nematics']:
        # by default plot as cells
        logger.debug('plot as cells')
        image_plot = Img(os.path.join(TA_path,'cells.tif')).astype(np.uint64)
    elif plot_type == 'bonds':
        logger.debug('plot as bonds')
        image_plot = Img(os.path.join(TA_path,'bonds.tif')).astype(np.uint64)
    elif plot_type == 'vertices':
        logger.debug('plot as vertices')
        image_plot = Img(os.path.join(TA_path,'vertices.tif')).astype(np.uint64)
    elif plot_type == 'packing':
        logger.debug('plot as packing')
        image_plot = Img(os.path.join(TA_path, 'cells.tif')).astype(np.uint64)
    else:
        logger.error('Plot type unknonw: \'' + str(plot_type) + '\'')
        return

    if erosion_dilation and plot_type in ['packing', 'cells', 'bonds', 'vertices']:
        s = generate_binary_structure(2, 1)
        if erosion_dilation > 0:
            # do gray dilation on image
            for _ in range(abs(erosion_dilation)):
                image_plot = grey_dilation(image_plot, footprint=s)
        elif erosion_dilation < 0:
            # do gray erosion on image
            for _ in range(abs(erosion_dilation)):
                image_plot = grey_erosion(image_plot, footprint=s)

    # then run sql command and see what TODO exactly
    # map all to the first
    # if full then find image in full list by its index maybe -> do that at some point

    database_path = os.path.join(TA_path, 'pyTA.db')
    # database_path = os.path.join(TA_path, 'TA.db')
    if not os.path.isfile(database_path):
        logger.error('pyTA db not found, nothing to plot ' + str(database_path))
        return

    query_results = None
    # db = None
    if db is None:
        try:
            db = TAsql(database_path)
            headers, query_results = db.run_SQL_command_and_get_results(SQL_command, return_header=True)

            # need get the max and min too of that command --> see how I can do that

        except:
            pass
        finally:
            if db is not None:
                db.close()
                db=None
    else:
        try:
            # use and query the master db but do not close it --> need query it twice to get the max and min of the values
            # db = TAsql(database_path)
            if SQL_command.strip().endswith(';'):
                SQL_command.replace(';','')

            # print('new sql command for masted db', SQL_command + ' WHERE frame_nb == '+str(current_frame))
            headers, query_results = db.run_SQL_command_and_get_results(SQL_command + ' WHERE frame_nb == '+str(current_frame), return_header=True)
        except:
            pass

    if query_results is None:
        logger.error('Plot failed because of an erroneous SQL command: ' + str(SQL_command))
        return

    # faut que je reparse le truc
    # ça marche mais voir comment faire ça de maniere portable en fait
    # print(query_results, headers)

    # TODO replace by panda dataframe to handle many more things or do smart loopings to read as if it were cols from the data --> in a way that is also not that hard
    method1 = False
    method2 = False
    if method1:
        # add a 0,0 mapping for cells --> TODO
        # in case of clones I may miss so many cells !!! --> think about the fastest way to add them !!!
        query_results.insert(0, (0, 0))

        # is there anything more secure ???
        # maybe yes with a loop but will be much less efficient --> think about it

        # ça marche mais si il manque plein de cellules il faut que je les ajoute --> si differe de continu --> soit ajouter des valeurs

        # j'en suis pas loin mais faut reflechir à la facon la plus facile de faire ça
        # need get the max of the image

        # convert this to a mapping array
        # should not be too hard to do
        formatted = np.asarray(query_results)
        # print(formatted.shape, formatted.dtype)  # 337 2

        # In[153]: keyarray = np.array(['S', 'M', 'L', 'XL'])
        # In[158]: data = np.array([[0, 2, 1], [1, 3, 2]])
        # In[159]: keyarray[data]
        # Out[159]:

        keyarray = formatted[:, 1]

        # output = np.zeros_like(image_plot)
        output = keyarray[image_plot]

        # print('output', output.shape, output.dtype)

        # convert left values to that in right --> indexing

        # no do some magic plotting
        # and somehow apply a lut ???

        # do stuff for the cells --> such as a plot of the area, etc...
        # need get global max and min --> a good idea and very easy todo with sqlite

        # create an empty image that we can then fill

        # can easily replace that by values from a database

        # for plot in rps:
        #     # do stuff
        #     # cell_output[cell.coords[:, 0], cell.coords[:, 1]] = 255
        #     output[plot.coords[:, 0], plot.coords[:, 1]] = plot.area # easy area plot in fact

        # if opacity ... do stuff
        # if Lut do further stuff

        # todo --> DO A SECOND MAPPING FOR LUT --> TODO -->

        # maybe could use negative values to get excluded cells
        # the color coding would be better with me !!! in fact

        # plt.imshow(output, cmap='Pastel1')
        # # plt.imshow(output, cmap='rainbow')
        # # plt.imshow(output, cmap='viridis')
        # # plt.imshow(output, cmap='gray')
        # plt.show()
    elif method2:
        # marche pas car l'erosion peut faire perdre des IDs de cellules --> need a fix
        rps = regionprops(image_plot)
        formatted = np.asarray(query_results, dtype=object)
        output = np.zeros_like(image_plot)

        # can easily replace that by values from a database
        # another way of doing, it might ultimately be simpler

        for iii, id in enumerate(formatted[:, 0]):
            # for plot in rps:
            plot = rps[int(id) - 1]
            # do stuff
            # cell_output[cell.coords[:, 0], cell.coords[:, 1]] = 255



            output[plot.coords[:, 0], plot.coords[:, 1]] = formatted[iii, 1]  # easy area plot in fact

        # # plt.imshow(output, cmap='Pastel1')
        # # plt.imshow(output, cmap='rainbow')
        # plt.imshow(output, cmap='viridis')
        # # plt.imshow(output, cmap='gray')
        # plt.show()
    else:
        # ça marche aussi même avec l'erosion --> pas le plus propre ni le plus efficace mais super robuste --> donc probablement utiliser ça !!!
        # rps = regionprops(image_plot)
        # in fact that still does not work great --> cannot handle None
        formatted = np.asarray(query_results, dtype=object) # required for handling complex outputs that contain string Nones and ints and floats

        # print(formatted) # all ok here --> why not plotting properly

        # print('formatted.shape', formatted.shape)


        # print(formatted, '\n test sqdqdsq', db==None, SQL_command)

        # do an autoscale for polarity based on the value of polarity for the max Q1 and Q2 values and at some point offer a rescaling factor maybe
        # print formatted
        # plot as cells
        if plot_type in ['packing', 'cells', 'bonds', 'vertices']:
            output = np.zeros_like(image_plot, dtype=float) # super important to have float here otherwise loses data !!!
            if return_mask:
                mask = np.zeros_like(image_plot)
            # can easily replace that by values from a database
            # another way of doing, it might ultimately be simpler

            # print(formatted[:, 0])
            # if not formatted:
            #     return




            for iii, id in enumerate(formatted[:, 0]):
                # for plot in rps:
                # plot = rps[int(id) - 1]
                # do stuff
                # cell_output[cell.coords[:, 0], cell.coords[:, 1]] = 255
                # output[plot.coords[:, 0], plot.coords[:, 1]] = formatted[iii, 1]  # easy area plot in fact

                # print(id, formatted[iii, 1], type(id)) # numpy.str_ --> for id --> why is that ??????
                idx = image_plot == id
                output[idx] = formatted[iii, 1]  # easy area plot in fact




                # print(formatted[iii, 1])

                if return_mask:
                    mask[idx] = 255
                # FutureWarning: elementwise comparison failed; returning scalar instead, but in the future will perform elementwise comparison
                #   output[image_plot == id] = formatted[iii, 1]  # easy area plot in fact

                # print(formatted[iii, 1]) # --> values are ok but I have a bug of conversion somewhere --> if values are very small  --> less than 1 they are rounded
            # plt.imshow(output)
            # plt.show()
            # up until here all is ok --> bug is in color coding
        elif plot_type.startswith('nematic'):
            # if return_mask:
            #     mask = np.zeros_like(image_plot)

            rps = regionprops(image_plot)
            # output = np.zeros_like(image_plot.shape, dtype=np.uint8)
            # im = numpy_to_PIL(output)

            # nb image is defined by with then height --> opposite of numpy!!!!
            # create a black image of the size of the image

            im = np.zeros_like(image_plot)
            # im = Image.new('RGB', (image_plot.shape[1],image_plot.shape[0]), (0,0,0))
            # draw = ImageDraw.Draw(im)

            has_scaling_factor = 'SCALING_FACTOR' in headers
            scaling_factor_idx = -1
            if has_scaling_factor:
                scaling_factor_idx = headers.index('SCALING_FACTOR')
                # print('scaling_factor_idx',scaling_factor_idx)
            has_colors = 'COLOR' in headers
            colors_idx = -1
            if has_colors:
                colors_idx= headers.index('COLOR')
            has_stroke_size = 'STROKE_SIZE' in headers
            stroke_idx = -1
            if has_stroke_size:
                stroke_idx =  headers.index('STROKE_SIZE')

            lines2ds = []
            # TODO --> implement nematic plot maybe with autoscaling depending on magnitude --> should not be too hard to do!!!
            for iii, id in enumerate(formatted[:, 0]):
                # need draw a line over the cell centroid
                # I need the rps corresponding to the cell to get its centroid
                # if
                plot = rps[int(id) - 1]
                # do stuff
                # cell_output[cell.coords[:, 0], cell.coords[:, 1]] = 255
                # output[plot.coords[:, 0], plot.coords[:, 1]] = formatted[iii, 1]  # easy area plot in fact
                centroid = plot.centroid


                # print(plot.centroid)
                # draw a nematic at this level
                # need

                # print('formatted[iii,1],formatted[iii,2]', formatted[iii,1],formatted[iii,2], formatted[iii,1]+1)



                nemat = Nematic(S1=formatted[iii, 1],S2=formatted[iii, 2], center = centroid)
                # nemat.draw(output)
                # nemat.draw(draw, stroke=2 if not has_stroke_size else formatted[iii, stroke_idx], rescaling_factor=None if not has_scaling_factor else formatted[iii, scaling_factor_idx], color= 0xFFFF00 if not has_colors else formatted[iii, colors_idx]) # todo scale it -> try that


                # TODO FINALIZE THE PLOTS WITH THE COLORS AND STUFF ALIKE




                l2d = nemat.toLine2D(rescaling_factor=1 if not has_scaling_factor else formatted[iii, scaling_factor_idx])
                # TODO add rescaling and line stroke
                if l2d is not None:
                    l2d.stroke = 2 if not has_stroke_size else formatted[iii, stroke_idx]
                    l2d.color =0xFFFF00 if not has_colors else formatted[iii, colors_idx]
                    lines2ds.append(l2d)

                    # print(l2d)


                # print(nemat)

                # would need to draw it as a line at some point but ok for now!!!
                # nemat

            im2d = Image2D(im)
            im2d.annotation.extend(lines2ds)

            im = im2d.save(None)
            im = im2d.convert_qimage_to_numpy(im)
            output = im

            # print('lines2ds', lines2ds)
            # output = PIL_to_numpy(im)



            if return_mask:
                mask = np.zeros_like(output)
                mask[output!=0]=255

            # print(output.shape)
            #
            # plt.imshow(output)
            # plt.show()

    # random_lut=random.choice(list(luts.keys()))
    # print('random.lut',random_lut)
    # plot.imshow(output)
    # plt.show()

    # print(output)


    # just get data normally and it is only for min and max that I will need the masterdb max and min
    # only for the LUT that I need that the rest can stay as such in fact
    if plot_type not in ['nematic','nematics']:
        # all seems ok --> just compare to TA to avoid issues
        # random_lut = 'HI_LO'
        lutcreator = PaletteCreator()
        luts = lutcreator.list
        # lut = lutcreator.create3(luts[lut])
        try:
            palette = lutcreator.create3(luts[lut])
        except:
            if lut is not None:
                logger.error('could not load the specified lut (' +str(lut)+') a gray lut is loaded instead') # --> ignoring or shall I default to gray
            # palette = None
            # default to grey palette
            palette = lutcreator.create3(luts['GRAY'])


        # maybe in the end also only get the modified pixels and plot only those --> especially for overlay or use transparency value to hide what should not be plotted --> set to fully transparent --> TODO!!!

        # lut = lutcreator.create3(luts['HI_LO'])

        # vraiment j'en suis pas loin --> juste voir comment finaliser les trucs
        # peut etre convertir l'output du truc de maniere simple
        # --> est-ce slow --> probablement pas
        # see how I can do that



        # ça marche presque --> juste tt finaliser maintenant !!!
        # output = apply_lut(output, lut, True, max=1000, min=900)
        # output = apply_lut(output, lut, True, max=6000, min=0) # max = 3*+ que le real max --> should all appear weak
        # output = apply_lut(output, lut, True, max=6000, min=100) # max = 3*+ que le real max --> should all appear weak # it seems to work but compare to TA
        # output = apply_lut(output, lut, True, max=3000, min=0) # 2*plus
        # is the bottom incorrect --> maybe
        # output = apply_lut(output, lut, True, max=1600, min=400) # 2*plus #a bit below real max # does it create a bug ???? --> maybe --> check it --> it looks weird to me
        # output = apply_lut(output, lut, True, max=0.9*output.max(), min=output.min()+(10.*(output.max()-output.min())/100.)) # 2*plus #a bit below real max # does it create a bug ???? --> maybe --> check it --> it looks weird to me


        # print(output) # no it's not rescaled
        # plt.imshow(output)
        # plt.show()

        # print(output.max())

        if plot_type == 'packing':
            # the trick to plot as packing is to maintain the data as it normally is and just bound it between 0 and 255 --> ignore all possible min and max
            # TODO see if I should not also add another value or other plots for example for bonds
            # or maybe plot as cell_id or alike and same stuff for bonds --> in such case apply no rescale whatever the data is... --> keep the raw values

            output = apply_lut(output, palette, True, min=0, max=255)
        else:

            # here I should get the max and min of the stuff to plot --> assumes there is a single value --> will that always be true????


            # need get real max and min here to apply a real global color code --> TODO


            # need get local or global max and min
            #
            # if SQL_command.strip().endswith(';'):
            #     SQL_command.replace(';', '')
            #
            # print(SQL_command + ' WHERE frame_nb == ' + str(current_frame))
            # headers, query_results = db.run_SQL_command_and_get_results(
            #     SQL_command + ' WHERE frame_nb == ' + str(current_frame), return_header=True)
            # probably need rerun the query

            # rerun the query

            min = None
            max = None
            # if db is
            if db is None:
                # print('getting local max and min TODO')
                try:
                    min, max = sort_col_numpy(formatted[:, -1], freq=freq)
                    logger.info('min is set to '+str(min)+ ' max is set to '+str(max))
                    # min is set to 1 max is set to 1 --> why
                except:
                    traceback.print_exc()
                    logger.error('local min/max would not be retrieved')

                # print()
            else:
                # print('getting global max and min TODO')
                # db.get_min_max()

                # print(formatted[:,-1]) # these are just the last values I would like to sort
                # seems ok now try to sort them

                # marche mais faut voir comment faire
                # if freq is not None:
                # is there any case where I should not do that ???

                headers, query_results = db.run_SQL_command_and_get_results(SQL_command, return_header=True)
                formatted = np.asarray(query_results, dtype=object)


                # need rerun the query without the frame nb

                # print(db.sort_col_numpy(formatted[:,-1]))

                # print(db.sort_col_numpy(formatted[:,-1], freq=(0.05, 0.05))) # ça marche pas mal enb fait
                try:
                    min, max = sort_col_numpy(formatted[:,-1], freq=freq)
                    logger.info('min is set to ' + str(min) + ' max is set to ' + str(max))
                except:
                    traceback.print_exc()
                    logger.error('global min/max would not be retrieved')

                # print(formatted.shape)


            output = apply_lut(output, palette, True, min=min, max=max) # 2*plus #a bit below real max # does it create a bug ???? --> maybe --> check it --> it looks weird to me

            # is that faster to get global max and min once for all

        # need check



        # je crois que tt marche mais faudrait vraiment tester en grandeur reelle
        # voir comment recuperer que les regions à vraiment plotter --> doit pas etre trop dur --> ignorer les trucs qui n'existent pas


        # if len(output.shape)==2:
        #     # plt.imshow(output, cmap='Pastel1')
        #     # plt.imshow(output, cmap='rainbow')
        #     plt.imshow(output, cmap='viridis')
        #     # plt.imshow(output, cmap='gray')
        # else:
        #     plt.imshow(output)
        # plt.show()
    if return_mask:
        if invert_mask:
            mask=Img.invert(mask)
        # also we apply the mask to the output
        output[mask!=0]=0
        return mask, output
    return output


def plot_as_image_old(plot_type='cells'):
    if plot_type == 'cells':
        # by default plot as cells
        logger.debug('plot as cells')
        image_plot = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012/cells.tif').astype(
            np.uint64)

    elif plot_type == 'bonds':
        logger.debug('plot as bonds')
        image_plot = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012/bonds.tif').astype(
            np.uint64)
    elif plot_type == 'vertices':
        logger.debug('plot as vertices')
        image_plot = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012/vertices.tif').astype(
            np.uint64)
    else:
        logger.error('Plot type unknonw: \'' + str(plot_type) + '\'')
        return

    rps = regionprops(image_plot)

    # do stuff for the cells --> such as a plot of the area, etc...
    # need get global max and min --> a good idea and very easy todo with sqlite

    # create an empty image that we can then fill
    output = np.zeros_like(image_plot)

    # can easily replace that by values from a database

    for plot in rps:
        # do stuff
        # cell_output[cell.coords[:, 0], cell.coords[:, 1]] = 255
        output[plot.coords[:, 0], plot.coords[:, 1]] = plot.area  # easy area plot in fact

    plt.imshow(output)
    plt.show()
    # plt.show(block=True)


# --> this is all very good and simple
# could also make graphs easily now --> TODO too


# TO plot a cell I need a command that generate a local id as first col then the value to plot as second col then various parameters that can be used for plotting -> try finalize that!!!!


if __name__ == '__main__':

    print(matplotlib.get_backend())
    matplotlib.validate_backend(matplotlib.get_backend())
    matplotlib.use('TkAgg')  # dirty fix for erroneous display with Qt5Agg


    if True:
        file = '/E/Sample_images/trash/complete/focused_Series010.png'
        # file = '/E/Sample_images/sample_images_pyta/surface_projection/210219.lif_t000.tif'
        # SQL_command = "SELECT local_id_cells,P1_polarity_ch2, P2_polarity_ch2, '#FF0000' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells"   #for TA db
        # SQL_command = "SELECT local_id,nb_of_vertices_or_neighbours FROM cells_2D WHERE is_border == FALSE"  # a bit more complex just to see which method to choose --> very easy
        SQL_command = "SELECT local_id,S1, S2, 10 AS SCALING_FACTOR FROM cells_2D"  # a bit more complex just to see which method to choose --> very easy
        # extras = {'LUT': 'DNA'}
        # extras = {'LUT': 'None'}
        # mask, output = plot_as_any(file, SQL_command, plot_type='cells', return_mask=True, **extras)
        # mask, output = plot_as_any(file, SQL_command, plot_type='cells', return_mask=True, **extras)
        # pb cannot plot is border --> somehow cannot plot booleans --> need a special treatment to avoid issues ??? assume min is 0 then and max is another thing ??? --> shall I implement that or a plot as boolean
        mask, output = plot_as_any(file, SQL_command, plot_type='nematics', return_mask=True)
        # --> bug
        if len(output.shape) == 2:
            # plt.imshow(output, cmap='Pastel1')
            # plt.imshow(output, cmap='rainbow')
            plt.imshow(output, cmap='viridis')
            # plt.imshow(output, cmap='gray')
        else:
            plt.imshow(output)
            # plt.imshow(mask)
        plt.show()

    if False:
        # master db plot
        master_db = createMasterDB(loadlist('/E/Sample_images/sample_images_pyta/surface_projection/list.lst'))


        SQL_command = "SELECT local_id, area FROM cells_2D"  # a bit more complex just to see which method to choose --> very easy
        extras = {'LUT': 'DNA'}
        file = '/E/Sample_images/sample_images_pyta/surface_projection/210219.lif_t000.tif'
        mask, output = plot_as_any(file, SQL_command, plot_type='cells', return_mask=True, db=master_db, current_frame=0, **extras)
        if len(output.shape) == 2:
            # plt.imshow(output, cmap='Pastel1')
            # plt.imshow(output, cmap='rainbow')
            plt.imshow(output, cmap='viridis')
            # plt.imshow(output, cmap='gray')
        else:
            plt.imshow(output)
            # plt.imshow(mask)
        plt.show()
        master_db.close()



    if False:
        file = '/E/Sample_images/sample_images_PA/trash_test_mem/mini_different_nb_of_channels/focused_Series012.png'
        # file = '/E/Sample_images/sample_images_pyta/surface_projection/210219.lif_t000.tif'
        # SQL_command = "SELECT local_id_cells,P1_polarity_ch2, P2_polarity_ch2, '#FF0000' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells"   #for TA db
        # SQL_command = "SELECT local_id,nb_of_vertices_or_neighbours FROM cells_2D WHERE is_border == FALSE"  # a bit more complex just to see which method to choose --> very easy
        SQL_command = "SELECT local_id, is_border FROM cells_2D"  # a bit more complex just to see which method to choose --> very easy
        # extras = {'LUT': 'DNA'}
        # extras = {'LUT': 'None'}
        # mask, output = plot_as_any(file, SQL_command, plot_type='cells', return_mask=True, **extras)
        # mask, output = plot_as_any(file, SQL_command, plot_type='cells', return_mask=True, **extras)
        # pb cannot plot is border --> somehow cannot plot booleans --> need a special treatment to avoid issues ??? assume min is 0 then and max is another thing ??? --> shall I implement that or a plot as boolean
        mask, output = plot_as_any(file, SQL_command, plot_type='cells', return_mask=True)
        # --> bug
        if len(output.shape) == 2:
            # plt.imshow(output, cmap='Pastel1')
            # plt.imshow(output, cmap='rainbow')
            plt.imshow(output, cmap='viridis')
            # plt.imshow(output, cmap='gray')
        else:
            plt.imshow(output)
            # plt.imshow(mask)
        plt.show()

    if False:
        file = '/E/Sample_images/sample_images_PA/trash_test_mem/mini_different_nb_of_channels/focused_Series012.png'
        file = '/E/Sample_images/sample_images_pyta/surface_projection/210219.lif_t000.tif'
        # SQL_command = "SELECT local_id_cells,P1_polarity_ch2, P2_polarity_ch2, '#FF0000' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells"   #for TA db
        # SQL_command = "SELECT local_id,nb_of_vertices_or_neighbours FROM cells_2D WHERE is_border == FALSE"  # a bit more complex just to see which method to choose --> very easy
        SQL_command = "SELECT local_id, nb_of_vertices_or_neighbours FROM cells_2D"  # a bit more complex just to see which method to choose --> very easy
        extras = {'LUT': 'PACKING'}
        # mask, output = plot_as_any(file, SQL_command, plot_type='cells', return_mask=True, **extras)
        mask, output = plot_as_any(file, SQL_command, plot_type='packing', return_mask=True, **extras)
        if len(output.shape) == 2:
            # plt.imshow(output, cmap='Pastel1')
            # plt.imshow(output, cmap='rainbow')
            plt.imshow(output, cmap='viridis')
            # plt.imshow(output, cmap='gray')
        else:
            plt.imshow(output)
            # plt.imshow(mask)
        plt.show()

    if True:

        # file = '/E/Sample_images/sample_images_PA/trash_test_mem/mini_different_nb_of_channels/focused_Series014.png' # to see if the nematic bug is linked to too many channels --> maybe the case
        # SQL_command = "SELECT local_id,Q1_polarity_ch0, Q2_polarity_ch0, '#FF0000' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells_2D WHERE is_border == FALSE"  # a bit more complex just to see which method to choose --> very easy

        file = '/E/Sample_images/sample_images_PA/trash_test_mem/mini_different_nb_of_channels/focused_Series012.png'
        # SQL_command = "SELECT local_id_cells,P1_polarity_ch2, P2_polarity_ch2, '#FF0000' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells"   #for TA db
        SQL_command = "SELECT local_id,Q1_polarity_ch1, Q2_polarity_ch1, '#FF0000' AS COLOR, 2 AS STROKE_SIZE, 0.1 AS SCALING_FACTOR FROM cells_2D WHERE is_border == FALSE"  # a bit more complex just to see which method to choose --> very easy


        # file = '/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012.png'
        # SQL_command = "PLOT AS CELLS SELECT local_id, area FROM cells_2D + LUT DNA + OPACITY 35% + DILATATION 1"
        # SQL_command = "PLOT AS NEMATICS SELECT center_x_cells,center_y_cells,P1_polarity_ch1, P2_polarity_ch1, '#FF0000' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells WHERE is_border=='false'"
        # SQL_command = "SELECT local_id, area FROM cells_2D"
        # SQL_command = "SELECT local_id, area FROM cells_2D WHERE is_border == FALSE"  # a bit more complex just to see which method to choose --> very easy
        # try plot a command that has a
        # SQL_command = "SELECT local_id, area FROM cells_2D WHERE is_border == FALSE"  # a bit more complex just to see which method to choose --> very easy
        # SQL_command = "SELECT local_id,Q1_polarity_ch1, Q2_polarity_ch1, '#FF0000' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells_2D WHERE is_border == FALSE"  # a bit more complex just to see which method to choose --> very easy


        # extras = {} # see how I can apply LUTS
        # extras = {'DILATION':1} # ça marche
        # extras = {'EROSION': 1}  # ça marche pas avec methode2 # seems to give me a bug --> why ??? --> shifts ids --> yes makes sense because some cells may be lost
        # extras = {'EROSION': 3}  # ça marche # seems to give me a bug --> why ??? --> shifts ids --> yes makes sense because some cells may be lost
        # extras = {'EROSION': 5}  # ça marche meme avec une strong erosion # seems to give me a bug --> why ??? --> shifts ids --> yes makes sense because some cells may be lost
        # extras = {'EROSION': 10}  # ça marche meme avec une strong erosion # seems to give me a bug --> why ??? --> shifts ids --> yes makes sense because some cells may be lost
        # extras = {'DILATION':22} # ça marche # dilation gives crappy results --> dialtion should be 1 not really more, just to fill the holes
        extras = {'LUT':'DNA'} # ça marche # dilation gives crappy results --> dialtion should be 1 not really more, just to fill the holes
        # extras = {'LUT':'copper'} # a cmap lut just to try # ça marche # dilation gives crappy results --> dialtion should be 1 not really more, just to fill the holes
        # output = plot_as_any(file, SQL_command, **extras)
        # output = plot_as_any(file, SQL_command, plot_type='cells',**extras)
        mask, output = plot_as_any(file, SQL_command, plot_type='nematics', return_mask=True,**extras)
        if len(output.shape)==2:
            # plt.imshow(output, cmap='Pastel1')
            # plt.imshow(output, cmap='rainbow')
            plt.imshow(output, cmap='viridis')
            # plt.imshow(output, cmap='gray')
        else:
            plt.imshow(output)
            # plt.imshow(mask)
        plt.show()

    # test all the stuff
    # TODO

    '''

    here are plenty of command line examples just to see how I can handle that
     titleNCommands.put("SQL command to select the 10 biggest cells (biggest area) (this will not generate a plot)", "SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells FROM cells ORDER BY area_cells DESC LIMIT 10");
            titleNCommands.put("plots 10 biggest cells (biggest area)", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells FROM cells ORDER BY area_cells DESC LIMIT 10");
            titleNCommands.put("plots 10 biggest cells (biggest area) and color code them by area (color code only applies to the current image, it is local not global)", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells FROM cells ORDER BY area_cells DESC LIMIT 10 + LUT DNA");
            titleNCommands.put("plots 10 biggest cells (biggest area) and color code them by area (use same color code/scaling parameters for all images)", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells FROM cells ORDER BY area_cells DESC LIMIT 10 + LUT DNA MIN=GLOBAL_MIN MAX=GLOBAL_MAX");
            titleNCommands.put("plots 10 biggest cells (biggest area) and color them in red", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells, '#FF0000' AS COLOR FROM cells ORDER BY area_cells DESC LIMIT 10");
            titleNCommands.put("plots 10 biggest cells (biggest area) and color them in red and plot them transparently (please select a background image)", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells, '#FF0000' AS COLOR FROM cells ORDER BY area_cells DESC LIMIT 10 + OPACITY 35%");
            titleNCommands.put("plots 10 biggest cells (biggest area) and color them in red, dilate them by one pixel and plot them transparently (please select a background image)", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells, '#FF0000' AS COLOR FROM cells ORDER BY area_cells DESC LIMIT 10 + DILATATION 1 + OPACITY 35%");
            titleNCommands.put("plots in red the elongation/stretch nematic scaled by a factor = 60, border cells are excluded", "PLOT AS NEMATICS SELECT center_x_cells,center_y_cells,S1_stretch_cells, S2_stretch_cells, '#FF0000' AS COLOR, '2' AS STROKE_SIZE, 60 AS SCALING_FACTOR FROM cells WHERE is_border_cell=='false'");
            titleNCommands.put("plots in red the polarity nematic of the image channel 1 scaled by a factor = 0.06, border cells are excluded, NB: will not display anything if the channel is empty", "PLOT AS NEMATICS SELECT center_x_cells,center_y_cells,P1_polarity_ch1, P2_polarity_ch1, '#FF0000' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells WHERE is_border_cell=='false'");
            titleNCommands.put("plots in green the polarity nematic of the image channel 2 scaled by a factor = 0.06, border cells are excluded, NB: will not display anything if the channel is empty", "PLOT AS NEMATICS SELECT center_x_cells,center_y_cells,P1_polarity_ch2, P2_polarity_ch2, '#00FF00' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells WHERE is_border_cell=='false'");
            titleNCommands.put("plots in blue the polarity nematic of the image channel 3 scaled by a factor = 0.06, border cells are excluded, NB: will not display anything if the channel is empty", "PLOT AS NEMATICS SELECT center_x_cells,center_y_cells,P1_polarity_ch3, P2_polarity_ch3, '#0000FF' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells WHERE is_border_cell=='false'");
            titleNCommands.put("plots in purple the polarity nematic of the 16 bits (single channel) image scaled by a factor = 0.0006, border cells are excluded", "PLOT AS NEMATICS SELECT center_x_cells,center_y_cells,P1_polarity_12bits, P2_polarity_12bits, '#FF00FF' AS COLOR, '2' AS STROKE_SIZE, 0.0006 AS SCALING_FACTOR FROM cells WHERE is_border_cell=='false'");
            titleNCommands.put("plots in red the intensity normalized polarity nematic of the image channel 1 scaled by a factor = 0.06, border cells are excluded, NB: will not display anything if the channel is empty", "PLOT AS NEMATICS SELECT center_x_cells,center_y_cells,P1_polarity_ch1_normalized, P2_polarity_ch1_normalized, '#FF0000' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells WHERE is_border_cell=='false'");
            titleNCommands.put("plots in green the intensity normalized polarity nematic of the image channel 2 scaled by a factor = 0.06, border cells are excluded, NB: will not display anything if the channel is empty", "PLOT AS NEMATICS SELECT center_x_cells,center_y_cells,P1_polarity_ch2_normalized, P2_polarity_ch2_normalized, '#00FF00' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells WHERE is_border_cell=='false'");
            titleNCommands.put("plots in blue the intensity normalized polarity nematic of the image channel 3 scaled by a factor = 0.06, border cells are excluded, NB: will not display anything if the channel is empty", "PLOT AS NEMATICS SELECT center_x_cells,center_y_cells,P1_polarity_ch3_normalized, P2_polarity_ch3_normalized, '#0000FF' AS COLOR, '2' AS STROKE_SIZE, 0.06 AS SCALING_FACTOR FROM cells WHERE is_border_cell=='false'");
            titleNCommands.put("plots in purple the intensity normalized polarity nematic of the 16 bits (single channel) image scaled by a factor = 0.0006, border cells are excluded", "PLOT AS NEMATICS SELECT center_x_cells,center_y_cells,P1_polarity_12bits_normalized, P2_polarity_12bits_normalized, '#FF00FF' AS COLOR, '2' AS STROKE_SIZE, 0.0006 AS SCALING_FACTOR FROM cells WHERE is_border_cell=='false'");
            titleNCommands.put("plots (-10, -5) yellow vectors centered on cells with a scaling factor 2, border cells are excluded", "PLOT AS VECTORS SELECT center_x_cells,center_y_cells, -10, -5, '#FFFF00' AS COLOR, '2' AS STROKE_SIZE, 2 AS SCALING_FACTOR FROM cells WHERE is_border_cell=='false'");
            //do the same for polarity for one channel in particular
            //maybe put all three channels separately
            titleNCommands.put("plots a line joining the centroid of cells and their first encountered pixel", "PLOT AS LINES SELECT first_pixel_x_cells,first_pixel_y_cells , center_x_cells,center_y_cells, '#0000FF' AS COLOR, '12' AS STROKE_SIZE, '0.3' AS STROKE_OPACITY FROM cells");
            titleNCommands.put("Plot cell geometry, triangles are in cyan, 4-sided green, 5-sided yellow, hexagons gray, 7-sided blue, cells 8 or 9 neighbours are red", "PLOT AS CELLS SELECT first_pixel_x_cells, first_pixel_y_cells, nb_of_vertices_cut_off FROM cells + LUT 3:#00FFFF 4:#00FF00 5:#FFFF00 6:#AAAAAA 7:#0000FF 8:#FF0000 9:#FF0000");
            titleNCommands.put("plots an arrow joining the centroid of cells and the first encountered cell pixel", "PLOT AS ARROWS SELECT center_x_cells,center_y_cells,first_pixel_x_cells , first_pixel_y_cells FROM cells");
            titleNCommands.put("plots an arrow joining the centroid of cells and the first encountered cell pixel, the arrow width is defined to 20 pixels and the arrowheight to 10", "PLOT AS ARROWS SELECT center_x_cells,center_y_cells,first_pixel_x_cells , first_pixel_y_cells, 20 AS ARROWHEAD_WIDTH, 10 AS ARROWHEAD_HEIGHT  FROM cells");
            titleNCommands.put("plots an half headed arrow (up) joining the centroid of cells and the first encountered cell pixel", "PLOT AS ARROWS SELECT center_x_cells,center_y_cells,first_pixel_x_cells , first_pixel_y_cells, 'HALF_HEAD_UP_ARROW' AS ARROWHEAD_TYPE FROM cells");
            titleNCommands.put("plots an half headed arrow (down) joining the centroid of cells and the first encountered cell pixel", "PLOT AS ARROWS SELECT center_x_cells,center_y_cells,first_pixel_x_cells , first_pixel_y_cells, 'HALF_HEAD_DOWN_ARROW' AS ARROWHEAD_TYPE FROM cells");
            titleNCommands.put("plots an double headed arrow joining the centroid of cells and the first encountered cell pixel", "PLOT AS ARROWS SELECT center_x_cells,center_y_cells,first_pixel_x_cells , first_pixel_y_cells, 'DOUBLE_HEAD_ARROW' AS ARROWHEAD_TYPE FROM cells");
            titleNCommands.put("plots an inhibition arrow joining the centroid of cells and the first encountered cell pixel", "PLOT AS ARROWS SELECT center_x_cells,center_y_cells,first_pixel_x_cells , first_pixel_y_cells, 'INIBITION_ARROW' AS ARROWHEAD_TYPE FROM cells");
            titleNCommands.put("plots a double headed inhibition arrow joining the centroid of cells and the first encountered cell pixel", "PLOT AS ARROWS SELECT center_x_cells,center_y_cells,first_pixel_x_cells , first_pixel_y_cells, 'DOUBLE_HEADED_INIBITION' AS ARROWHEAD_TYPE FROM cells");

            titleNCommands.put("plots a text containing the 'local id of cells' (i.e. their cell nb) centered over the cell centroid", "PLOT AS STRINGS SELECT center_x_cells, center_y_cells , local_id_cells, '#00FF00' AS COLOR, '#FF0000' AS BG_COLOR, '22' AS FONT_SIZE, 'ITALIC' AS FONT_STYLE FROM cells");
            titleNCommands.put("plots a square the width of which is proportional to the cell area/100 where the top left corner of the square is the centroid", "PLOT AS SQUARES SELECT center_x_cells, center_y_cells  , area_cells/100, '8016' AS COLOR,'30000' AS FILL_COLOR,'0.3' AS FILL_OPACITY FROM cells");
            //titleNCommands.put("plots a vector joining the first encountered pixel of the cell to its cenroid", "PLOT AS VECTORS SELECT first_pixel_x_cells,first_pixel_y_cells , center_x_cells,center_y_cells, '#00FF00' AS COLOR, '1' AS STROKE_SIZE, 'DEFAULT_ARROW' AS ARROWHEAD_TYPE, '5' AS ARROWHEAD_HEIGHT,'3' AS ARROWHEAD_HEIGHT, '0.3' AS STROKE_OPACITY FROM cells;");
            titleNCommands.put("plots all cells with an area between 1000 and 1200 pixels. NB may not show anything depending on your cell area", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, '#FFFF00' AS COLOR FROM cells WHERE area_cells > 1000 AND area_cells <1200  + OPACITY 35% + DILATATION 1");
            titleNCommands.put("color codes all cells according to their area", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, area_cells FROM cells + LUT DNA");
            titleNCommands.put("color codes (according to cell area) a vector joining the first encountered pixel of the cell to its cenroid", "PLOT AS ARROWS SELECT first_pixel_x_cells,first_pixel_y_cells , center_x_cells,center_y_cells, area_cells AS COLOR, '1' AS STROKE_SIZE, 'DEFAULT_ARROW' AS ARROWHEAD_TYPE, '5' AS ARROWHEAD_HEIGHT,'3' AS ARROWHEAD_HEIGHT, '0.3' AS STROKE_OPACITY FROM cells + LUT DNA");
            titleNCommands.put("color codes bonds according to their orientation (ignores bonds consisting only of two vertices)", "PLOT AS BONDS SELECT first_pixel_x_bonds, first_pixel_y_bonds, bond_orientation FROM bonds WHERE first_pixel_y_bonds != 'null' + LUT DNA");
            titleNCommands.put("draws cells as polygons (using vertex coordinates)", "PLOT AS POLYGONS SELECT vx_coords_cells, area_cells AS COLOR,area_cells AS FILL_COLOR,  '0.65' AS STROKE_SIZE FROM cells + LUT DNA + OPACITY 35%");
            titleNCommands.put("Test a SQL command (i.e. preview table output)", "SELECT * FROM CELLS");
            titleNCommands.put("color code cells accoding to their elongation", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, S0_stretch_cells FROM cells + LUT DNA + OPACITY 35% + DILATATION 1");
            titleNCommands.put("color code cells, except border cells, accoding to their elongation", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, S0_stretch_cells FROM cells WHERE is_border_cell=='false' + LUT DNA + OPACITY 35% + DILATATION 1");
            titleNCommands.put("color code cells, except border cells and cells immediately adjacent to border cells, accoding to their elongation", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, S0_stretch_cells FROM cells WHERE is_border_cell_plus_one=='false' + LUT DNA + OPACITY 35% + DILATATION 1");

            titleNCommands.put("(Global) color code cells belonging to clone #22 according to their amount of stretch (NB: virtual clone #22 must exist for the command to work)", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, S0_stretch_cells FROM cells NATURAL JOIN tracked_clone_022 + LUT DNA MIN=GLOBAL_MIN MAX=GLOBAL_MAX + OPACITY 35% + DILATATION 1");
            titleNCommands.put("(Global) color code cells belonging to clone #0 according to their amount of stretch (NB: virtual clone #0 must exist for the command to work)", "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, S0_stretch_cells FROM cells NATURAL JOIN tracked_clone + LUT DNA MIN=GLOBAL_MIN MAX=GLOBAL_MAX + OPACITY 35% + DILATATION 1");


    '''

    if False:
        plot_SQL('PLOT     AS  CELLS SELECT * FROM CELLS;')
        plot_SQL('PLOT     AS            BONDS SELECT * FROM CELLS;')
        plot_SQL('PLOT     AS  VERTices                      SELECT * FROM CELLS;')
        plot_SQL('PLOT     AS TEST SELECT * FROM CELLS;')
        plot_SQL('PLOT AS        POLYGON SELECT * FROM CELLS;')

        plot_SQL(
            "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells FROM cells ORDER BY area_cells DESC LIMIT 10")  # --> see how I can do that
        plot_SQL(
            "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells FROM cells ORDER BY area_cells DESC LIMIT 10 + LUT DNA")  # --> see how I can do that
        plot_SQL(
            "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells, '#FF0000' AS COLOR FROM cells ORDER BY area_cells DESC LIMIT 10 + DILATATION 1 + OPACITY 35%")  # in fact each extra is starting with a + --> quite easy to get
        # --> ignore pluses in extras
        plot_SQL(
            "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells FROM cells ORDER BY area_cells DESC LIMIT 10 + LUT DNA MIN=GLOBAL_MIN MAX=GLOBAL_MAX")  # --> maybe not that hard because the values if complex are connected to MIn and MAX --> if all are like that then I can parse all extras as keys and values and all is gonna be a piece of cake and easy to detect and report errors --> TODO

        plot_SQL(
            "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, S0_stretch_cells FROM cells + LUT DNA + OPACITY 35% + DILATATION 1")  # --> maybe not that hard because the values if complex are connected to MIn and MAX --> if all are like that then I can parse all extras as keys and values and all is gonna be a piece of cake and easy to detect and report errors --> TODO
        plot_SQL(
            "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, S0_stretch_cells FROM cells + LUT DNA + OPACITY 35% + DILATATION 1+BOUGA BOUU")  # --> maybe not that hard because the values if complex are connected to MIn and MAX --> if all are like that then I can parse all extras as keys and values and all is gonna be a piece of cake and easy to detect and report errors --> TODO
        plot_SQL(
            "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, S0_stretch_cells FROM cells + LUT DNA + OPACITY 35% + DILATATION 1+BOUGA")  # --> maybe not that hard because the values if complex are connected to MIn and MAX --> if all are like that then I can parse all extras as keys and values and all is gonna be a piece of cake and easy to detect and report errors --> TODO
        plot_SQL(
            "PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells, S0_stretch_cells FROM cells + LUT DNA + TRANSPARENCY 35% + DILATATION 1+BOUGA")  # --> maybe not that hard because the values if complex are connected to MIn and MAX --> if all are like that then I can parse all extras as keys and values and all is gonna be a piece of cake and easy to detect and report errors --> TODO
        plot_SQL(
            "PLOT AS CELLS SELECT first_pixel_x_cells, first_pixel_y_cells, nb_of_vertices_cut_off FROM cells + LUT 3:#00FFFF 4:#00FF00 5:#FFFF00 6:#AAAAAA 7:#0000FF 8:#FF0000 9:#FF0000")  # --> maybe not that hard because the values if complex are connected to MIn and MAX --> if all are like that then I can parse all extras as keys and values and all is gonna be a piece of cake and easy to detect and report errors --> TODO

        # value and how to do that

        # or parse the extras between the pluses and accept all the data in between is data connected to the first keyword --> TODO
        # _strip_from_command("PLOT AS CELLS SELECT first_pixel_x_cells,first_pixel_y_cells , area_cells FROM cells ORDER BY area_cells DESC LIMIT 10", "PLOT AS CELLS")
        # --> ok that seems to work

        print(parse_opacity("OPACITY 35%".split()))

        # for _ in range(abs(0)):
        #     print('hello')

    if False:
        plot_as_image_old(plot_type=None)
        plot_as_image_old()



# here I will put numpy tips
import matplotlib.pyplot as plt
import numpy as np

def get_value_from_nd_coords(img, nd_coords):
    return img[convert_nd_coords_to_numpy_format(nd_coords)]

def convert_nd_coords_to_numpy_format(nd_coords):
    return tuple(nd_coords.T)

def set_coords_to_value(img, nd_coords, value):
    img[convert_nd_coords_to_numpy_format(nd_coords)]=value

def convert_numpy_bbox_to_coord_pairs(bbox):
    '''
    returns coords pairs from a linear bbox (1,2,3,200,600,150) -->
    [[  1 200]
    [  2 600]
    [  3 150]]
    :param bbox: typically a region.bbox from regionprops
    :return: coord pairs
    '''
    if not isinstance(bbox, np.ndarray):
        bbox = np.asarray(bbox)
    return np.reshape(bbox, (2, int(bbox.size / 2))).T


def get_image_bounds(original_image):
    bounds = []
    for dim in range(len(original_image.shape)):
        bounds.append(0)
        bounds.append(original_image.shape[dim])

    # bounds = convert_numpy_bbox_to_coord_pairs(bounds)
    bounds = np.asarray(bounds).reshape(-1, 2)
    return bounds


def filter_nan_rows(coords):
    if coords.size == 0:
        print('empty array --> retruning array')
        return coords
    return coords[~np.isnan(coords).any(axis=1)]

if __name__ == '__main__':
    test = np.zeros(shape=(32,32))
    coords = np.asarray([[0,0],[16,16],[20,20]])
    set_coords_to_value(test, coords, 255) # --> same as test[coords[:,0],coords[:,1]] # and what is great is that infinite nb of dims are supported --> can write generic code for 2D and 3D --> really cool --> replace everywhere in my code
    plt.imshow(test)
    plt.show()




import math
from skimage.measure import label, regionprops
from timeit import default_timer as timer
import numpy as np
import matplotlib.pyplot as plt


def add_2d_border(image2d):
    insertHere = (slice(1, image2d.shape[0] - 1), slice(1, image2d.shape[1] - 1))
    cells_with_borders = np.zeros_like(image2d)
    cells_with_borders.fill(0)

    cells_with_borders[insertHere] = image2d[insertHere]
    # plt.imshow(cells_with_borders)
    # plt.show()
    # print(insertHere)
    image2d = cells_with_borders

    # print(image2d.shape)
    return image2d


def create_horizontal_gradient(cells):
    cells = add_2d_border(cells)
    horiz_gradient = np.zeros_like(cells)

    for j in range(0, cells.shape[-2], 1):
        counter = 1
        last_id = 0
        for i in range(0, cells.shape[-1], 1):
            # dark stuff or beginning of line creates a beginning of the gradient and a counter
            # then try go back by n pixels when new white/black line is detected

            current_cell_id = cells[j][i]

            if current_cell_id != 0 and current_cell_id != last_id:
                last_id = current_cell_id

            if current_cell_id == 0 and last_id != 0:
                last_id = 0
                # divide counter by 2
                half_counter = int(counter / 2.)
                counter = 1  # restart counter
                # found a new cell edge --> go backwards to create the gradient
                reverse_counter = 1
                for ii in range(0, -half_counter + 1, -1):
                    horiz_gradient[j][i + ii] = reverse_counter
                    reverse_counter += 1

            if current_cell_id != 0 and current_cell_id == last_id:
                horiz_gradient[j][i] = counter
                counter += 1
    return horiz_gradient


def create_vertical_gradient(cells):
    cells = add_2d_border(cells)
    vertical_gradient = np.zeros_like(cells)

    for i in range(0, cells.shape[-1], 1):
        counter = 1
        last_id = 0
        for j in range(0, cells.shape[-2], 1):
            current_cell_id = cells[j][i]

            if current_cell_id != 0 and current_cell_id != last_id:
                last_id = current_cell_id

            if current_cell_id == 0 and last_id != 0:
                last_id = 0
                # divide counter by 2
                half_counter = int(counter / 2.)
                counter = 1  # restart counter
                # found a new cell edge --> go backwards to create the gradient
                reverse_counter = 1
                for jj in range(0, -half_counter + 1, -1):
                    vertical_gradient[j + jj][i] = reverse_counter
                    reverse_counter += 1

            if current_cell_id != 0 and current_cell_id == last_id:
                vertical_gradient[j][i] = counter
                counter += 1
    return vertical_gradient


def get_seeds(cells, one_seed_per_cell=True):
    # TODO really need make sure there is only one seed per cell --> if many then reduce to one by keeping only the biggest --> good idea and may improve things
    # really worth a test...
    horiz_gradient = create_horizontal_gradient(cells)
    vertical_gradient = create_vertical_gradient(cells)

    # plt.imshow(horiz_gradient)
    # plt.show()
    #
    # plt.imshow(vertical_gradient)
    # plt.show()

    combined_gradients_ready_for_wshed, seeds = _get_seeds(cells, horiz_gradient, vertical_gradient)

    # plt.imshow(combined_gradients_ready_for_wshed)
    # plt.show()

    # Img(combined_gradients_ready_for_wshed, dimensions='hw').save('/E/Sample_images/sample_images_epiguy_pyta/images_with_different_bits/predict/gradient.tif')

    # if there are several seeds for a cell then just keep the biggest --> good idea
    # how can I do that --> maybe simply do so by counting ids for stuff and remove smallest
    # count how many seeds are found for each cell

    if one_seed_per_cell:
        new_seeds = label(seeds.astype(np.uint8), connectivity=1, background=0)
        props_seeds = regionprops(new_seeds)

        extra_seeds_to_remove = []

        for region in regionprops(cells):
            # remove small seeds
            cells_found = []
            for coordinates in region.coords:
                cells_found.append(new_seeds[coordinates[0], coordinates[1]])
            cells_found = list(dict.fromkeys(cells_found))
            if 0 in cells_found:
                cells_found.remove(0)
            if len(cells_found) > 1:
                # extra_seeds_to_remove
                # loop over seeds by area and keep only the best/biggest
                # see
                max_area = 0
                for cell in cells_found:
                    region = props_seeds[cell - 1]
                    max_area = max(max_area, region.area)

                for cell in cells_found:
                    if not props_seeds[cell - 1].area == max_area:
                        extra_seeds_to_remove.append(cell)

        for region in props_seeds:
            if region.label in extra_seeds_to_remove:
                for coordinates in region.coords:
                    seeds[coordinates[0], coordinates[1]] = 0  # do remove the seed

    return combined_gradients_ready_for_wshed, seeds


def _get_seeds(cells, horiz_gradient, vertical_gradient):
    # combine gradients
    combined_gradients_ready_for_wshed = horiz_gradient + vertical_gradient

    highest_pixels = np.zeros_like(cells)

    for region in regionprops(cells, intensity_image=combined_gradients_ready_for_wshed):
        max_val = 0

        factor = math.sqrt(
            region.area) * 30 / 100  # try scale seed by cell size --> use area for that # ideally should have jut one seed per cell and scale with size

        for j in range(0, region.intensity_image.shape[-2], 1):
            for i in range(0, region.intensity_image.shape[-1], 1):
                if region.intensity_image[j, i] >= max_val:
                    max_val = region.intensity_image[j, i]

        # region.image[...] = False  # do it better with fill
        region.image.fill(False)

        region.image[region.intensity_image >= (max_val - factor)] = True

        for coordinates in region.coords:
            highest_pixels[coordinates[0], coordinates[1]] = 255
    return combined_gradients_ready_for_wshed, highest_pixels


if __name__ == '__main__':
    start = timer()



# faire deux modes --> un in mem
# faire une table 'properties' si elle existe


# TODO lui faire aussi charger une liste d'images et tt un tas de fichiers db --> à faire
# peut etre faire une table style image properties
# je pourrais reinjecter

# cols to have --> time and d/x ratio and  maybe also the xyz px sizes and the unit of size
# see how and where I can have that done
# maybe also set it as a parameter
# can I allow copy below if 1 or None
# maybe if not known set it to none then it is ignored
# then allow plots
# ce truc pourrait etre genere avant meme de rentrer la dedans à l'etape du preprocess --> good idea and no need to modify the file which is a bit painful

# ça marche mais faut le rendre un peu plus smart

import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import numpy as np
import sqlite3
# import pandas as pd
# from qtpy import QtSql
# from qtpy.QtCore import Qt
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QWidget, QApplication, QTableWidgetItem, QPushButton, QVBoxLayout

from epyseg.ta.database.sql_column_name_dialog import SQL_column_name_and_type
from epyseg.utils.loadlist import loadlist, smart_TA_list
from epyseg.ta.database.my_qt_table import MyTableWidget
from epyseg.ta.database.sql import TAsql, get_properties_master_db, \
    reinject_properties_to_TA_files


class Example(QWidget):
    def __init__(self, table_name='properties', lst=None, db_to_connect_to=None, non_editable_columns=[0,1]):
        super().__init__()
        self.db_connect = None
        self.non_editable_columns = non_editable_columns
        self.lst = lst
        self.table_name = table_name
        self.initUI()
        self.set_db(db_to_connect_to)
        # self.update_required = False
        self.rows_to_update = []


    def initUI(self):
        hbox = QVBoxLayout(self)

        self.table_widget = MyTableWidget()
        self.table_widget.itemChanged.connect(self.item_changed)

        # db_connect = QtSql.QSqlDatabase.addDatabase("QSQLITE")
        # db_connect.setDatabaseName("exemple.db")
        # db_connect.open()
        # model = QtSql.QSqlQueryModel(parent=self)
        # model.setQuery("SELECT * FROM stocks")
        # tableView.setModel(model)
        #
        # db_connect.close()

        # model = QtGui.QStandardItemModel()

        # print(len(cursor_obj.fetchall()))

        # if self.db_connect is not None:

        # faudrait creer une hbox où je sauve le reste --> TODO but ok for now

        hbox.addWidget(self.table_widget)

        save_button = QPushButton('Save')
        save_button.clicked.connect(self.reinject_modified_properties)
        hbox.addWidget(save_button)

        # self.tab1c.layout = QVBoxLayout()
        self.add_column_to_properties_table_button = QPushButton("Add column")
        self.add_column_to_properties_table_button.clicked.connect(self.add_column_to_properties_table)
        # self.tab1c.layout.addWidget(self.add_column_to_properties_table_button)
        # self.tab1c.setLayout(self.tab1c.layout)
        hbox.addWidget(self.add_column_to_properties_table_button)



        self.setLayout(hbox)

        self.show()

    def get_column_names(self):
        col_names = []
        for i in range(self.table_widget.columnCount()):
            header_item = self.table_widget.horizontalHeaderItem(i)
            col_names.append(header_item.text())
        return col_names

    def add_column_to_properties_table(self):
        # ask the user for new col name and type
        colnmame_n_type, ok = SQL_column_name_and_type.get_value(parent=self,title="New Column Name", existing_column_name=self.get_column_names())
        if ok:
            db = TAsql(filename_or_connection=self.db_connect)
            # We add a column with the given type to the current table
            db.add_column(self.table_name,colnmame_n_type[0],col_type=colnmame_n_type[1]) # maybe format string so that it is nice looking # shall i also specify type of column such as number or string or ignore ???
            self.update_db()
            # Force update all rows! because we have added a new column
            for rrr in range(self.table_widget.rowCount()):
                self.rows_to_update.append(rrr)

    def reinject_modified_properties(self):
        self.reinject_updated_props()

    def set_db(self, db):
        # if a db was already opened before then close it
        self.close()
        # if self.db_connect is not None:
        #     self.db_connect.close()

        if isinstance(db, list):
            self.lst = db
            if db is None or not db:
                db = None
            else:
                db = get_properties_master_db(db)

        if db is None:
            self.db_connect = None
            self.cursor = None
            self.table_widget.clear()
            return

        if isinstance(db, str):
            self.db_connect = sqlite3.connect(db)
            self.cursor = self.db_connect.cursor()
        else:
                self.db_connect =db.con
                self.cursor = db.cur

        self.update_db()


    def update_db(self):
        if self.db_connect is None:
            return

        # print('self.table_name', self.table_name)
        # self.db_connect = db

        # rowcount = cursor.execute('''SELECT COUNT(*) FROM '''+table_name).fetchone()[0]
        #

        # print(rowcount)
        self.cursor.execute('''SELECT * FROM ''' + self.table_name)
        content = self.cursor.fetchall()

        self.table_widget.setRowCount(len(content))
        self.table_widget.setColumnCount(len((content[0])))

        column_names = [description[0] for description in self.cursor.description]

        # print(cursor.rowcount)

        # print(cursor.c)
        # tableView.setColumnCount(cursor.) # --> ok --> needs be set too
        # tableView.setColumnCount(2) # --> ok --> needs be set too

        # for row, form in enumerate(cursor):
        #     for column, item in enumerate(form):
        #         self.tblTable.setItem(row, column, QtGui.QTableWidgetItem(str(item)))
        # first = True

        self.table_widget.itemChanged.disconnect()

        for row, form in enumerate(content):
            # tableView.insertRow(row)

            # print(len(form))
            # if first:
            #     first = False
            #     tableView.setColumnCount(len(form))
            #     tableView.setRowCount(len(cursor))

            for column, item in enumerate(form):
                # print(row, column, str(item))# this is ok
                itm = QTableWidgetItem(str(item))
                # we make unsaved TA only columns not editable to avoid issues
                if self.non_editable_columns is not None and self.non_editable_columns:
                    if column in self.non_editable_columns:
                        itm.setFlags(itm.flags() ^ Qt.ItemIsEditable)
                self.table_widget.setItem(row, column,itm)

        # db_connect.close()
        # tableView.update()

        # need set all column names and headers
        self.table_widget.setHorizontalHeaderLabels(column_names)

        self.table_widget.itemChanged.connect(self.item_changed)
        # self.update_required = False # reset update required
        self.rows_to_update=[]
        # print('oubsi', column_names)

    def item_changed(self, item):
        # self.update_required = True
        # print('something changed --> update required')
        row = item.row()
        # col = item.column()
        self.rows_to_update.append(row)
        # print(self.rows_to_update) # rather use this because it also allows me to save just the necessary stuff --> no need to update all in fact

    # TODO --> do that more properly some day
    def saveToDb(self, tablename):
        table_data = {}
        for i in range(self.table_widget.columnCount()):
            items = []
            for j in range(self.table_widget.rowCount()):
                item = self.table_widget.item(j, i)
                items.append(item.text())
            header_item = self.table_widget.horizontalHeaderItem(i)
            # n_column = str(i) if h_item is None else h_item.text()
            # print(h_item.text())
            table_data[header_item.text()] = items

        # df = pd.DataFrame(data=d)
        # engine = sqlite3.connect(db_filename)
        # df.to_sql(tablename, con=self.db_connect,if_exists='replace')# pas top car ajoute une colonne et fait une conversion de type

        # print(d)
        self.dict_to_sql(table_data, tablename)

    def dict_to_sql(self, values, tablename):
        # columns = ', '.join(values.keys())
        # placeholders = ', '.join('?' * len(values))

        # print(columns, placeholders)

        # what I want is not insert but update all
        # sql = 'INSERT INTO stocks ({}) VALUES ({})'.format(columns, placeholders)
        # sql = 'UPDATE stocks ({}) VALUES ({})'.format(columns, placeholders)
        # sql = ''' UPDATE stocks
        #           SET ({}) VALUES ({})
        #           WHERE id = ?'''.format(columns, placeholders)

        # sql = 'UPDATE table_name SET column1 = value1, column2 = value2...., columnN = valueN WHERE [condition];'


        # values = [int(x) if isinstance(x, bool) else x for x in values.values()]

        # print(values)
        # print(sql)

        # bug because of too many entries

        # for iii in range(len(values[0])):
        #     row = [x[iii] for x in values]
        #     self.cursor.execute(sql, row)

        # keys = values.keys()
        # for value in values.values():

        columns =' = ? ,\n\t'.join(values.keys())+' = ? '
        # print(columns)

        # in fact need update row by row --> see how

        # purchases = [('2006-03-28', 'BUY', 'IBM', 1000, 45.00),
        #              ('2006-04-05', 'BUY', 'MSFT', 1000, 72.00),
        #              ('2006-04-06', 'SELL', 'IBM', 500, 53.00),
        #              ]
        # c.executemany('INSERT INTO stocks VALUES (?,?,?,?,?)', purchases)

        datas = list(values.values())
        # print(datas)
        # print(datas)

        rowid = [iii+1 for iii in range(len(datas[0]))]
        # datas = np.asarray(datas)
        # print(datas.shape)
        # print(rowid)
        # datas = np.vstack([datas, rowid])
        # datas = np.append(datas, rowid)
        datas.append(rowid)
        # print(np.array(datas))
        # ok but needs be reordered into a format that is more friendly --> that takes
        datas = np.asarray(datas).T.tolist()
        # print('-->',datas)
        # update all at once maybe
        # for iii, data in enumerate(datas):
        #     data.append(iii+1)

        # for iii,val in enumerate(datas):

        sql = "UPDATE "+tablename+" \nSET " +columns +"\nWHERE ROWID = ?"
        # sql ='''UPDATE Database_Name.Table_Name
        #         SET Column1_Name = value1, Column2_Name = value2,...
        #         WHERE condition'''
        # val = #("Valley 345", "Canyon 123")

        # print(sql)
        # print(val)

        # self.cursor.execute(sql, val)
        self.cursor.executemany(sql, datas)

        self.db_connect.commit()
        # self.cursor.executemany(sql, values)

    def close(self):
        try:
            if self.db_connect is not None:
                self.db_connect.close()
                self.db_connect = None
        except:
            pass

    def reinject_updated_props(self):
        if self.rows_to_update:
                # lst = self.get_full_list(warn_on_empty_list=False)
                if self.lst is not None and self.lst:
                    self.saveToDb('properties')
                    # self.properties_table.update_required = False
                    # I need get its db and reinject it
                    reinject_properties_to_TA_files(self.lst, TAsql(filename_or_connection=self.db_connect), indices_to_update=self.rows_to_update) # TODO rinject just the necessary lines
                # else maybe ask where to save ???
        self.rows_to_update = []
        # self.properties_table.close()

#
# db_connect = sqlite3.connect("database.db")
# cur = db_connect.cursor()
# cur.execute("SELECT id, num, data FROM table")
# content = cur.fetchall()
# row_count = len(content)
# db_connect.commit()
# cur.close()
# db_connect.close()
# model = QtGui.QStandardItemModel()
#
# lst1 = []
# lst2 = []
# lst3 = []
#
# for i in content:
# lst1.append(i[0])
# lst2.append(i[1])
# lst3.append(i[2])
#
# for i in range(0, row_count):
# item1 = QtGui.QStandardItem(lst1[i])
# item2 = QtGui.QStandardItem(lst2[i])
# item3 = QtGui.QStandardItem(lst3[i])
#
# model.appendRow([item1, item2, item3])
# item1.setEditable(False)
# window.tableView.setModel(model)


def _tmp_remove_props(lst):
    database_list = smart_TA_list(lst, 'pyTA.db')
    for db_file in database_list:
        db = TAsql(db_file)
        db.drop_table('properties2')
        db.close()


# TODO connect this to the main db et creer le truc


if __name__ == "__main__":
    # if True:
    #     db_connect = sqlite3.connect('test.db')
    #     print(type(db_connect))
    #     import sys
    #     sys.exit(0)



    app = QApplication([])

    lst = loadlist('/E/Sample_images/sample_images_PA/trash_test_mem/mini_different_nb_of_channels/list.lst')
    # _tmp_remove_props(lst)



    # db = get_properties_master_db(lst)

    ex = Example(table_name = 'properties')
    ex.set_db(lst)
    ex.show()
    app.exec_()

    # TODO do some code to reinject all the lines into each file of the list --> quite easy
    # just copy row per row and get all but the two first columns

    # in fact I need upate the master db first but ok for a test
    # ex.saveToDb('properties')
    # reinject_properties_to_TA_files(lst, db)

    ex.close()


# align objects left right top or bottom
# maybe just need methods and not even a class
# logger
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy.QtCore import QRectF

from epyseg.draw.shapes.point2d import Point2D
from epyseg.tools.logger import TA_logger
import numpy as np

logger = TA_logger()

# TODO replace all functions by that to avoid duplicated code

# def setToWidth(space, width_in_px, *objects):
#     if width_in_px is None:
#         return
#     bounds = updateBoudingRect(*objects)
#     left = Point2D(bounds.topLeft())
#     incompressible_width = 0
#     for img in objects:
#         incompressible_width+=img.getIncompressibleWidth()
#     # si row doit maintenir same height until reached desired width is reached --> change its width --> doit aussi connaitre la incompressible height des objects
#     pure_image_width = (bounds.width()) - incompressible_width
#     # print(width_in_px, self.getIncompressibleWidth(), (nb_cols - 1.) * self.space )
#     width_in_px -= incompressible_width
#     ratio = width_in_px / pure_image_width
#     for img in objects:
#         img.setToWidth(img.boundingRect().width() * ratio)
#     packX(space, left, *objects)
#     # self.updateBoudingRect()

# def setToHeight2(space, height_in_px, *objects):
#     if height_in_px is None:
#         return
#     bounds = updateBoudingRect(*objects)
#     left = Point2D(bounds.topLeft())
#     incompressible_height = 0
#     for img in objects:
#         incompressible_height+=img.getIncompressibleHeight()
#     # si row doit maintenir same height until reached desired width is reached --> change its width --> doit aussi connaitre la incompressible height des objects
#     pure_image_height = (bounds.height()) - incompressible_height
#     # print(width_in_px, self.getIncompressibleWidth(), (nb_cols - 1.) * self.space )
#     height_in_px -= incompressible_height
#     ratio = height_in_px / pure_image_height
#     for img in objects:
#         img.setToHeight(img.boundingRect().height() * ratio)
#     packY(space, left, *objects)
#     # self.updateBoudingRect()



#from timeit import default_timer as timer see why so slow --> see what part of computation is slow and remove it
def setToHeight(space, height_in_px, *objects):
    if height_in_px is None:
        return
    bounds = updateBoudingRect(*objects)
    left = Point2D(bounds.topLeft())
    for img in objects:
        img.setToHeight(height_in_px)
    packX(space, left, *objects)
    # self.updateBoudingRect()

def _brute_force_find_width(col1, min_h, max_h, increment, desired_height):
    # to be faster --> find sign inversion and stop there and take closest
    # stop on sign change in fact
    closest = None
    mn = 100000000
    old_sign = None
    break_soon = False


    for width in np.arange(min_h, max_h, increment):
        col1.setToWidth(width)
        # print('bob', height, col1.width(), col1.height())
        # in fact before that it becomes <0

        if old_sign is not None:
            if old_sign > 0 and col1.height() - desired_height < 0:
                # print('changing sign --> stopping')
                break_soon = True
            elif old_sign < 0 and col1.height() - desired_height > 0:
                # print('changing sign --> stopping')
                break_soon = True

        if old_sign is None:
            if col1.height() - desired_height < 0:
                old_sign = -1
            else:
                old_sign = 1
        if (col1.height() - desired_height) <= mn and col1.height() - desired_height >= 0:
            closest = (col1.width(), col1.height(), width)
            mn = col1.height() - desired_height
            if break_soon:
                # print('stopping at ',height)  # stopping at  41.14420004379321 min_h, max_h 40.14420004379321 43.19553376948639 # --> a vraiment stoppé vite en fait
                break

    # print('break_soon, closest, min_h, max_h, increment',break_soon, closest, min_h, max_h, increment)
    return closest


def _brute_force_find_height(col1, min_h, max_h, increment, desired_width):
        # to be faster --> find sign inversion and stop there and take closest
        # stop on sign change in fact
        closest = None
        mn = 100000000
        old_sign = None
        break_soon = False
        for height in np.arange(min_h, max_h, increment):
            col1.setToHeight(height)
            # print('bob', height, col1.width(), col1.height())
            # in fact before that it becomes <0

            if old_sign is not None:
                if old_sign > 0 and col1.width() - desired_width < 0:
                    # print('changing sign --> stopping')
                    break_soon = True
                elif old_sign < 0 and col1.width() - desired_width > 0:
                    # print('changing sign --> stopping')
                    break_soon = True

            if old_sign is None:
                if col1.width() - desired_width < 0:
                    old_sign = -1
                else:
                    old_sign = 1
            if (col1.width() - desired_width) <= mn and col1.width() - desired_width >= 0:
                closest = (col1.width(), col1.height(), height)
                mn = col1.width() - desired_width
                if break_soon:
                    # print('stopping at ',height)  # stopping at  41.14420004379321 min_h, max_h 40.14420004379321 43.19553376948639 # --> a vraiment stoppé vite en fait
                    break
        return closest

def setToWidth2(space, width_in_px, *objects):
    if width_in_px is None:
        return
    bounds = updateBoudingRect(*objects)
    left = Point2D(bounds.topLeft())
    for img in objects:
        img.setToWidth(width_in_px)
    packY(space, left, *objects)



def updateBoudingRect(*objects):
        '''updates the image bounding rect depending on content'''
        bounding_rect = QRectF()
        x = None
        y = None
        x2 = None
        y2 = None
        for img in objects:
            topLeft = img.get_P1()
            if x is None:
                x = topLeft.x()
            if y is None:
                y = topLeft.y()
            x = min(topLeft.x(), x)
            y = min(topLeft.y(), y)

            # print(img, img.boundingRect(), type(img))
            # print(img, img.boundingRect(), type(img), img.boundingRect().height())

            if x2 is None:
                x2 = topLeft.x() + img.boundingRect().width()
            if y2 is None:
                y2 = topLeft.y() + img.boundingRect().height()
            x2 = max(topLeft.x() + img.boundingRect().width(), x2)
            y2 = max(topLeft.y() + img.boundingRect().height(), y2)

        bounding_rect.setX(x)
        bounding_rect.setY(y)
        bounding_rect.setWidth(x2 - x)
        bounding_rect.setHeight(y2 - y)
        return bounding_rect


def packX(space=3, reference=None, *objects_to_align_with_respect_to_ref):
    skip_first = False
    if objects_to_align_with_respect_to_ref is None or len(objects_to_align_with_respect_to_ref) == 0:
        logger.warning("Nothing to align...")
        return
    if reference is None:
        if len(objects_to_align_with_respect_to_ref) < 2:
            logger.warning("Reference is None, can't align anything...")
            return
        else:
            reference = objects_to_align_with_respect_to_ref[0]
            skip_first = True
            # del objects_to_align_with_respect_to_ref[0]

    last_x = reference.get_P1().x()+reference.width()
    last_y = reference.get_P1().y()

    for i in range(len(objects_to_align_with_respect_to_ref)):
            if i == 0 and skip_first:
                continue
            img = objects_to_align_with_respect_to_ref[i]
            # print('begin', last_x)
            if i != 0:
                last_x += space


            # print('before', last_x, img.boundingRect(), space)
            img.set_P1(last_x, img.get_P1().y())
            # print(img.boundingRect(), last_x, img.get_P1().x())
            last_x = img.boundingRect().x() + img.boundingRect().width()
            # print('end', last_x)
    # self.updateBoudingRect()

    # align everything with respect to ref
    # get first point and align so that the all have the same x

ORIENTATION_TOP_TO_BOTTOM = 0
ORIENTATION_BOTTOM_TO_TOP = 1

# need do a code for two orientations --> from top to bottom and reverse --> can be useful to do...


def packYreverse(space=3, reference=None, *objects_to_align_with_respect_to_ref):
    skip_first = False
    if objects_to_align_with_respect_to_ref is None or len(objects_to_align_with_respect_to_ref) == 0:
        logger.warning("Nothing to align...")
        return
    if reference is None:
        if len(objects_to_align_with_respect_to_ref) < 2:
            logger.warning("Reference is None, can't align anything...")
            return
        else:
            reference = objects_to_align_with_respect_to_ref[0]
            skip_first = True
            # del objects_to_align_with_respect_to_ref[0]

    # last_x = reference.get_P1().x()
    last_y = reference.get_P1().y()

    for i in range(len(objects_to_align_with_respect_to_ref)):
        if i==0 and skip_first:
            continue
        img = objects_to_align_with_respect_to_ref[i]
        if i != 0:
            last_y -= space
            # last_y -= img.boundingRect().height()
        # print(last_y, img.get_P1())
        img.set_P1(img.get_P1().x(), last_y-img.boundingRect().height())
        # get all the bounding boxes and pack them with desired space in between
        # get first point and last point in x
        x = img.boundingRect().x()
        y = img.boundingRect().y()
        # last_x = img.boundingRect().x() + img.boundingRect().width()
        last_y = img.boundingRect().y()
    # self.updateBoudingRect()

def packY(space=3, reference=None, *objects_to_align_with_respect_to_ref):
    skip_first = False
    if objects_to_align_with_respect_to_ref is None or len(objects_to_align_with_respect_to_ref) == 0:
        logger.warning("Nothing to align...")
        return
    if reference is None:
        if len(objects_to_align_with_respect_to_ref) < 2:
            logger.warning("Reference is None, can't align anything...")
            return
        else:
            reference = objects_to_align_with_respect_to_ref[0]
            skip_first = True
            # del objects_to_align_with_respect_to_ref[0]

    # last_x = reference.get_P1().x()
    last_y = reference.get_P1().y() + reference.height()

    for i in range(len(objects_to_align_with_respect_to_ref)):
        if i==0 and skip_first:
            continue
        img = objects_to_align_with_respect_to_ref[i]
        if i != 0:
            last_y += space
        # print(last_y, img.get_P1())
        img.set_P1(img.get_P1().x(), last_y)
        # get all the bounding boxes and pack them with desired space in between
        # get first point and last point in x
        x = img.boundingRect().x()
        y = img.boundingRect().y()
        # last_x = img.boundingRect().x() + img.boundingRect().width()
        last_y = img.boundingRect().y() + img.boundingRect().height()
    # self.updateBoudingRect()


    # align everything with respect to ref
    # get first point and align so that the all have the same x


def alignLeft(reference=None, *objects_to_align_with_respect_to_ref):
    skip_first = False
    if objects_to_align_with_respect_to_ref is None or len(objects_to_align_with_respect_to_ref) == 0:
        logger.warning("Nothing to align...")
        return
    if reference is None:
        if len(objects_to_align_with_respect_to_ref) < 2:
            logger.warning("Reference is None, can't align anything...")
            return
        else:
            reference = objects_to_align_with_respect_to_ref[0]
            skip_first = True
            # del objects_to_align_with_respect_to_ref[0]

    last_x = reference.get_P1().x()
    # last_y = reference.get_P1().y()

    for i in range(len(objects_to_align_with_respect_to_ref)):
        if i==0 and skip_first:
            continue
        img = objects_to_align_with_respect_to_ref[i]

        img.set_P1(last_x, img.get_P1().y())
        # get all the bounding boxes and pack them with desired space in between
        # get first point and last point in x
        # x = img.boundingRect().x()
        # y = img.boundingRect().y()
        # last_x = img.boundingRect().x() + img.boundingRect().width()
        # last_y = img.boundingRect().y() + img.boundingRect().height()
    # self.updateBoudingRect()


    # align everything with respect to ref
    # get first point and align so that the all have the same x


def alignTop(reference=None, *objects_to_align_with_respect_to_ref):
    skip_first = False
    if objects_to_align_with_respect_to_ref is None or len(objects_to_align_with_respect_to_ref) == 0:
        logger.warning("Nothing to align...")
        return
    if reference is None:
        if len(objects_to_align_with_respect_to_ref) < 2:
            logger.warning("Reference is None, can't align anything...")
            return
        else:
            reference = objects_to_align_with_respect_to_ref[0]
            skip_first = True
            # del objects_to_align_with_respect_to_ref[0]

    # last_x = reference.get_P1().x()
    last_y = reference.get_P1().y()

    for i in range(len(objects_to_align_with_respect_to_ref)):
        if i==0 and skip_first:
            continue
        img = objects_to_align_with_respect_to_ref[i]

        img.set_P1(img.get_P1().x(), last_y)
        # get all the bounding boxes and pack them with desired space in between
        # get first point and last point in x
        # x = img.boundingRect().x()
        # y = img.boundingRect().y()
        # last_x = img.boundingRect().x() + img.boundingRect().width()
        # last_y = img.boundingRect().y() + img.boundingRect().height()
    # self.updateBoudingRect()


    # align everything with respect to ref
    # get first point and align so that the all have the same x


def alignBottom(reference=None, *objects_to_align_with_respect_to_ref):
    skip_first = False
    if objects_to_align_with_respect_to_ref is None or len(objects_to_align_with_respect_to_ref) == 0:
        logger.warning("Nothing to align...")
        return
    if reference is None:
        if len(objects_to_align_with_respect_to_ref) < 2:
            logger.warning("Reference is None, can't align anything...")
            return
        else:
            reference = objects_to_align_with_respect_to_ref[0]
            skip_first = True
            # del objects_to_align_with_respect_to_ref[0]

    # last_x = reference.get_P1().x()
    last_y = reference.get_P1().y()+reference.height() # should be the end point of every stuff

    for i in range(len(objects_to_align_with_respect_to_ref)):
        if i==0 and skip_first:
            continue
        img = objects_to_align_with_respect_to_ref[i]


        pos_y = img.get_P1().y()+img.height()
        difference_y = last_y-pos_y
        img.set_P1(img.get_P1().x(), img.get_P1().y()+difference_y)
        # get all the bounding boxes and pack them with desired space in between
        # get first point and last point in x
        # x = img.boundingRect().x()
        # y = img.boundingRect().y()
        # last_x = img.boundingRect().x() + img.boundingRect().width()
        # last_y = img.boundingRect().y() + img.boundingRect().height()
    # self.updateBoudingRect()


    # align everything with respect to ref
    # get first point and align so that the all have the same x


# bug here now with the images
def alignRight(reference=None, *objects_to_align_with_respect_to_ref):
    skip_first = False
    if objects_to_align_with_respect_to_ref is None or len(objects_to_align_with_respect_to_ref) == 0:
        logger.warning("Nothing to align...")
        return
    if reference is None:
        if len(objects_to_align_with_respect_to_ref) < 2:
            logger.warning("Reference is None, can't align anything...")
            return
        else:
            reference = objects_to_align_with_respect_to_ref[0]
            skip_first = True
            # del objects_to_align_with_respect_to_ref[0]

    last_x = reference.get_P1().x()+reference.width() # should be the end point of every stuff
    # last_y = reference.get_P1().y()+reference.height() # should be the end point of every stuff

    for i in range(len(objects_to_align_with_respect_to_ref)):
        if i==0 and skip_first:
            continue
        img = objects_to_align_with_respect_to_ref[i]

        pos_x = img.get_P1().x()+img.width()
        difference_x = last_x-pos_x
        img.set_P1(img.get_P1().x()+difference_x, img.get_P1().y())
        # get all the bounding boxes and pack them with desired space in between
        # get first point and last point in x
        # x = img.boundingRect().x()
        # y = img.boundingRect().y()
        # last_x = img.boundingRect().x() + img.boundingRect().width()
        # last_y = img.boundingRect().y() + img.boundingRect().height()
    # self.updateBoudingRect()


    # align everything with respect to ref
    # get first point and align so that the all have the same x


def alignCenterH(reference=None, *objects_to_align_with_respect_to_ref):
    skip_first = False
    if objects_to_align_with_respect_to_ref is None or len(objects_to_align_with_respect_to_ref) == 0:
        logger.warning("Nothing to align...")
        return
    if reference is None:
        if len(objects_to_align_with_respect_to_ref) < 2:
            logger.warning("Reference is None, can't align anything...")
            return
        else:
            reference = objects_to_align_with_respect_to_ref[0]
            skip_first = True
            # del objects_to_align_with_respect_to_ref[0]

    last_x = reference.get_P1().x()+reference.width()/2. # should be the end point of every stuff
    # last_y = reference.get_P1().y()+reference.height() # should be the end point of every stuff

    for i in range(len(objects_to_align_with_respect_to_ref)):
        if i==0 and skip_first:
            continue
        img = objects_to_align_with_respect_to_ref[i]

        pos_x = img.get_P1().x()+img.width()/2.
        difference_x = last_x-pos_x
        img.set_P1(img.get_P1().x()+difference_x, img.get_P1().y())
        # get all the bounding boxes and pack them with desired space in between
        # get first point and last point in x
        # x = img.boundingRect().x()
        # y = img.boundingRect().y()
        # last_x = img.boundingRect().x() + img.boundingRect().width()
        # last_y = img.boundingRect().y() + img.boundingRect().height()
    # self.updateBoudingRect()


    # align everything with respect to ref
    # get first point and align so that the all have the same x

def alignCenterV(reference=None, *objects_to_align_with_respect_to_ref):
    skip_first = False
    if objects_to_align_with_respect_to_ref is None or len(objects_to_align_with_respect_to_ref) == 0:
        logger.warning("Nothing to align...")
        return
    if reference is None:
        if len(objects_to_align_with_respect_to_ref) < 2:
            logger.warning("Reference is None, can't align anything...")
            return
        else:
            reference = objects_to_align_with_respect_to_ref[0]
            skip_first = True
            # del objects_to_align_with_respect_to_ref[0]

    # last_x = reference.get_P1().x()
    last_y = reference.get_P1().y()+reference.height()/2. # should be the end point of every stuff

    for i in range(len(objects_to_align_with_respect_to_ref)):
        if i==0 and skip_first:
            continue
        img = objects_to_align_with_respect_to_ref[i]


        pos_y = img.get_P1().y()+img.height()/2.
        difference_y = last_y-pos_y
        img.set_P1(img.get_P1().x(), img.get_P1().y()+difference_y)
        # get all the bounding boxes and pack them with desired space in between
        # get first point and last point in x
        # x = img.boundingRect().x()
        # y = img.boundingRect().y()
        # last_x = img.boundingRect().x() + img.boundingRect().width()
        # last_y = img.boundingRect().y() + img.boundingRect().height()
    # self.updateBoudingRect()


    # align everything with respect to ref
    # get first point and align so that the all have the same x
#
# def alignRight(reference=None, *objects_to_align_with_respect_to_ref):
#     skip_first = False
#     if objects_to_align_with_respect_to_ref is None or len(objects_to_align_with_respect_to_ref) == 0:
#         logger.warning("Nothing to align...")
#         return
#     if reference is None:
#         if len(objects_to_align_with_respect_to_ref) < 2:
#             logger.warning("Reference is None, can't align anything...")
#             return
#         else:
#             reference = objects_to_align_with_respect_to_ref[0]
#             skip_first = True
#             # del objects_to_align_with_respect_to_ref[0]
#
#     last_x = reference.get_P1().x()
#     # last_y = reference.get_P1().y()
#
#     for i in range(len(objects_to_align_with_respect_to_ref)):
#         if i==0 and skip_first:
#             continue
#         img = objects_to_align_with_respect_to_ref[i]
#
#         img.set_P1(last_x, img.get_P1().y())
#         # get all the bounding boxes and pack them with desired space in between
#         # get first point and last point in x
#         # x = img.boundingRect().x()
#         # y = img.boundingRect().y()
#         # last_x = img.boundingRect().x() + img.boundingRect().width()
#         # last_y = img.boundingRect().y() + img.boundingRect().height()
#     # self.updateBoudingRect()
#
#
#     # align everything with respect to ref
#     # get first point and align so that the all have the same x




# TODO do plots of the logs
from epyseg.img import Img
from epyseg.tools.logger import TA_logger
from timeit import default_timer as timer

logger = TA_logger()

import traceback
import tensorflow as tf
import os


class My_saver_callback(tf.keras.callbacks.Callback):
    '''a custom saver callback

    '''

    def __init__(self, save_output_name, deepTA, epochs=None, keep_n_best=5, output_folder_for_models=None,
                 monitor='loss', verbose=0, save_best_only=False, save_training_log=True, keep_best_validation=True,
                 save_weights_only=False, progress_callback=None):
        '''init saver callback

        Parameters
        ----------
        save_output_name : string
            generic name for model saving

        deepTA : deep learning tool


        epochs : int
            total nb of epochs (required to display progress)

        keep_n_best : int
            number of best models (lower loss) to keep during training

        output_folder_for_models : string
            folder where models should be saved

        monitor : str
            loss to monitor (could be loss or val_loss if validation set exists)

        verbose : int
            verbose level

        save_best_only : boolean
            if True only saves the best model (overrides keep_n_best)

        save_training_log :  boolean
            if True saves training log (containing losses and metrics outputs along with the epoch nb, useful for plots)

        keep_best_validation :  boolean
            keeps best model (lowest loss) on validation

        save_weights_only : boolean
            only save weights not model and optimizer

        progress_callback : anything to display progress
            dipslays progress

        '''

        self.save_weights_only = save_weights_only
        self.output_folder_for_models = output_folder_for_models
        self.save_output_name = save_output_name
        if self.output_folder_for_models is not None:
            if not os.path.exists(self.output_folder_for_models):
                os.makedirs(self.output_folder_for_models, exist_ok=True)
            self.save_output_name = os.path.join(self.output_folder_for_models, self.save_output_name)
        # print('joined path', self.save_output_name)
        self.deepTA = deepTA
        self.epochs = epochs
        self.keep_n_best = keep_n_best
        self.monitor = monitor
        self.save_best_only = save_best_only
        self.verbose = verbose
        self.save_training_log = save_training_log
        self.keep_best_validation = keep_best_validation
        self.best_val_loss = 10000
        self.progress_callback = progress_callback
        if self.save_best_only:
            self.keep_n_best = 1  # only keep the best
        self.stop_me = False
        self.best_model = None

    def on_train_begin(self, logs={}):
        self.best_loss = 1_000_000  # keeps the best/lowest loss value
        self.losses = []
        self.accs = []
        self.best_kept = []

        # self.rank_n_score={}
        # self.rank_n_filename = {}
        self.loss_n_filename = {}

        if self.save_training_log:
            # save raw log for plot purpose
            open(self.save_output_name + '.log', 'w+').close()

    def on_train_end(self, logs={}):
        return

    def on_epoch_begin(self, epoch, logs={}):
        self.start_time = timer()
        try:
            if self.progress_callback is not None and self.epochs is not None:
                self.progress_callback.emit((epoch / self.epochs) * 100)
            else:
                logger.info(str((epoch / self.epochs) * 100) + '%')
        except:
            pass
        return

    # in fact need store rename first then run stuff
    def get_save_name(self, epoch, value):
        return self.save_output_name + '-ep%04d-l%0.6f.h5' % (epoch + 1, value)

    def on_epoch_end(self, epoch, logs={}):
        try:
            end_time = timer()
            epoch_time = end_time - self.start_time
            total_time_in_hours = (epoch_time * (self.epochs - epoch)) / 3600
            if total_time_in_hours >= 1.:
                logger.info('Estimated remaining run time: ' + str(total_time_in_hours) + ' hour(s)')
            else:
                total_time_in_hours *= 60
                if total_time_in_hours >= 1.:
                    logger.info('Estimated remaining run time: ' + str(total_time_in_hours) + ' minute(s)')
                else:
                    total_time_in_hours *= 60
                    logger.info('Estimated remaining run time: ' + str(total_time_in_hours) + ' second(s)')
        except:
            pass

        self.losses.append(logs.get(self.monitor))
        logger.debug('logs ' + str(logs))

        # TODO remove that ???
        if self.keep_best_validation and not self.stop_me:
            val_loss = logs.get('val_loss')
            if val_loss is not None and val_loss < self.best_val_loss:
                self.best_val_loss = val_loss
                # always overwrites it...
                logger.debug('Saving best val loss model')
                # TODO maybe change this and replace by accuracy ????
                if not self.save_weights_only:
                    logger.info('saving file: ' + self.save_output_name + '_best_val_loss.h5')
                    self.model.save(self.save_output_name + '_best_val_loss.h5')
                else:
                    logger.info('saving file: ' + self.save_output_name + '_best_val_loss.h5')
                    self.model.save_weights(self.save_output_name + '_best_val_loss.h5')

        save_file_name = self.save_output_name + '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor))

        if self.keep_n_best is not None and not self.stop_me:
            # now seems ok but need do the same for all other steps
            # this is the part of the code I need to change to overwrite
            if self.best_kept:
                if len(self.best_kept) < self.keep_n_best:
                    self.best_kept.append(logs.get(self.monitor))
                    if logs.get(self.monitor) is not None:

                        self.best_kept.sort(reverse=True)
                        # print('sorted stuff', self.best_kept)
                        # print('idx=', self.best_kept.index(logs.get(self.monitor)))
                        # seems ok now need rename stuff

                        # self.best_kept.sort(reverse=True)
                        # if logs.get(self.monitor) is not None:
                        #     if logs.get(self.monitor) < self.best_kept[0]:
                        #         loss = self.best_kept[0]
                        #         self.best_kept[0] = logs.get(self.monitor)

                        # if not self.save_weights_only:
                        current_rank = self.best_kept.index(logs.get(self.monitor))
                        # print('need rename all images with rank <= ',current_rank)

                        real_rank = len(self.best_kept) - current_rank - 1
                        # print('new image real rank', real_rank)

                        save_file_name = self.save_output_name + '-' + str(
                            real_rank) + '.h5'  # + str('-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor))
                        # print('filename ', save_file_name)

                        # en fait need rename tt les trucs de valeur superieure

                        # print('current files',self.loss_n_filename, logs.get(self.monitor))
                        # print('max rank', self.keep_n_best)
                        # counter = 0
                        updates = {}
                        for val in self.best_kept:
                            # if val>=logs.get(self.monitor):

                            if val == logs.get(self.monitor):
                                break

                            # need rename ? to rank + 1
                            # if rank is superior to max then delete file

                            if not val in self.loss_n_filename:
                                continue
                            # print('need rename ', self.loss_n_filename[val], ' to current rank + 1') # should be easy

                            new_index = self.best_kept.index(val)
                            real_new_index = len(self.best_kept) - new_index - 1
                            # print('expected_new_index_for_the image', real_new_index)
                            save_file_name2 = self.save_output_name + '-' + str(real_new_index) + '.h5'
                            # print(os.path.exists(save_file_name2), self.loss_n_filename[val])

                            try:
                                # bug fix to prevent filling google drive bin
                                open(save_file_name2,
                                     'w').close()  # overwrite and make the file blank instead - ref: https://stackoverflow.com/a/4914288/3553367
                            except:
                                pass
                            # os.rename seems to cause a crash if file already exists on windows os see https://stackoverflow.com/questions/45636341/is-it-wise-to-remove-files-using-os-rename-instead-of-os-remove-in-python
                            os.replace(self.loss_n_filename[val], save_file_name2)

                            # print('new_name_for_the_file', save_file_name2, 'before', self.loss_n_filename[val])
                            # self.loss_n_filename[val] = save_file_name2
                            updates[val] = save_file_name2

                            # if counter == 0:
                            #     print(remove)
                            # counter+=1

                        # else:
                        #     print(val,'<',logs.get(self.monitor))

                        # print("updates",updates)
                        # print('orig', self.loss_n_filename)
                        self.loss_n_filename.update(updates)
                        # print('orig corrected', self.loss_n_filename)
                        # then should save current as rank

                        # save rank and name
                        self.loss_n_filename[logs.get(
                            self.monitor)] = save_file_name  # self.save_output_name + str('-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor)))
                        # print('loss n filename',self.loss_n_filename)

                        # IN FACT SIMPLE ALWAYS NEED POP FIRST IN THE LIST THE WAY IT'S INVERTED

                        # need also store top 5 files

                        # up until here should be ok and no action is required it's only after that action is require

                        # logger.info('saving file: ' + self.save_output_name + str(                                '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor))))
                        logger.info('saving file: ' + save_file_name)
                        # self.model.save(                                self.save_output_name + '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor)))
                        if not self.save_weights_only:
                            self.model.save(save_file_name)
                        else:
                            self.model.save_weights(save_file_name)
                        # every image of rank equal or > need be renamed
                        # save as rank_ --> so that one can easily cut name and replace it
                        # else:
                        #     logger.info('saving file: ' + self.save_output_name + str(
                        #         '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor))))
                        #     self.model.save_weights(
                        #         self.save_output_name + '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor)))
                else:
                    self.best_kept.sort(reverse=True)
                    if logs.get(self.monitor) is not None:
                        if logs.get(self.monitor) < self.best_kept[0]:
                            # remove image that already exists
                            self.loss_n_filename.pop(self.best_kept[0], None)
                            # loss = self.best_kept[0]
                            self.best_kept[0] = logs.get(self.monitor)

                            self.best_kept.sort(reverse=True)
                            current_rank = self.best_kept.index(logs.get(self.monitor))
                            real_rank = len(self.best_kept) - current_rank - 1
                            save_file_name = self.save_output_name + '-' + str(
                                real_rank) + '.h5'  # + str('-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor))

                            updates = {}
                            for val in self.best_kept:
                                if val == logs.get(self.monitor):
                                    break

                                if not val in self.loss_n_filename:
                                    continue

                                new_index = self.best_kept.index(val)
                                real_new_index = len(self.best_kept) - new_index - 1
                                save_file_name2 = self.save_output_name + '-' + str(real_new_index) + '.h5'

                                # print('tada',new_index, real_new_index, save_file_name2)
                                # print(val)
                                # print(os.path.exists(save_file_name2), os.path.exists(self.loss_n_filename[val]))

                                try:
                                    # bug fix to prevent filling google drive bin
                                    open(save_file_name2,
                                         'w').close()  # overwrite and make the file blank instead - ref: https://stackoverflow.com/a/4914288/3553367
                                except:
                                    pass
                                # os.rename seems to cause a crash if file already exists on windows os see https://stackoverflow.com/questions/45636341/is-it-wise-to-remove-files-using-os-rename-instead-of-os-remove-in-python

                                # print(self.loss_n_filename,self.loss_n_filename[val],save_file_name2)
                                # print(os.path.exists(save_file_name2), os.path.exists(self.loss_n_filename[val])) # the dest exists but not the source --> why is that ????


                                # somehow there is a random bug here but no clue why
                                try:
                                    os.replace(self.loss_n_filename[val], save_file_name2)
                                except:
                                    logger.error('this should never have been reached trying ultimate rescue solution')
                                    print('tada',new_index, real_new_index, save_file_name2)
                                    print(val)
                                    print(self.best_kept)
                                    print(updates)
                                    # print(os.path.exists(save_file_name2), os.path.exists(self.loss_n_filename[val]))
                                    print(self.loss_n_filename,self.loss_n_filename[val],save_file_name2)
                                    print(os.path.exists(save_file_name2), os.path.exists(self.loss_n_filename[val])) # the dest exists but not the source --> why is that ????
                                    #
                                    # somehow there is a random bug here but no clue why
                                    print('renaming 2', self.loss_n_filename[val], 'to', save_file_name2)
                                    traceback.print_exc()
                                    if not self.save_weights_only:
                                        self.model.save(save_file_name2)
                                    else:
                                        self.model.save_weights(save_file_name2)
                                    # try:
                                    #     os.replace(self.loss_n_filename[val], save_file_name2)
                                    # except:
                                    #     pass

                                # print('renaming 2', self.loss_n_filename[val], 'to', save_file_name2)

                                updates[val] = save_file_name2

                            # print("updates", updates)
                            # print('orig', self.loss_n_filename)
                            self.loss_n_filename.update(updates)
                            # print('orig corrected', self.loss_n_filename)

                            self.loss_n_filename[logs.get(
                                self.monitor)] = save_file_name  # self.save_output_name + str('-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor)))
                            # print('loss n filename', self.loss_n_filename)

                            logger.info('saving file: ' + save_file_name)
                            if not self.save_weights_only:
                                self.model.save(save_file_name)
                            else:
                                self.model.save_weights(save_file_name)

                            # logger.info('saving file: ' + self.save_output_name + str(
                            #     '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor))))
                            # if not self.save_weights_only:
                            #
                            #     self.model.save(
                            #         self.save_output_name + '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor)))
                            # else:
                            #     self.model.save_weights(
                            #         self.save_output_name + '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor)))
                            # no need to remove any longer...
                            # logger.debug('deleting file: ' + str(self.save_output_name + '-ep*-l%0.6f.h5' % (loss)))
                            # fileList = glob.glob(self.save_output_name + '-ep*-l%0.6f.h5' % (loss))
                            # for f in fileList:
                            #     os.remove(f)
            else:
                self.best_kept.append(logs.get(self.monitor))
                if logs.get(self.monitor) is not None and not self.stop_me:
                    # self.loss_n_filename[logs.get(self.monitor)] = self.save_output_name + str(                        '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor))) # tada
                    save_file_name = self.save_output_name + '-' + str(0) + '.h5'
                    self.loss_n_filename[logs.get(self.monitor)] = save_file_name
                    logger.info('saving file: ' + save_file_name)
                    if not self.save_weights_only:
                        # logger.info('saving file: ' + self.save_output_name + str(                            '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor))))
                        # self.model.save(                            self.save_output_name + '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor)))
                        self.model.save(save_file_name)
                    else:
                        # logger.info('saving file: ' + self.save_output_name + str(                            '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor))))
                        # self.model.save_weights(                            self.save_output_name + '-ep%04d-l%0.6f.h5' % (epoch + 1, logs.get(self.monitor)))
                        self.model.save_weights(save_file_name)

            f = open(self.save_output_name + '_losses_n_corresponding_files' + '.log', "w")
            f.write(str(self.loss_n_filename))
            f.close()



        else:
            # save all files with their loss
            if logs.get(self.monitor) is not None and not self.stop_me:
                logger.info('saving file: ' + save_file_name)
                if not self.save_weights_only:
                    self.model.save(save_file_name)
                else:
                    self.model.save_weights(save_file_name)

        # best loss is always #1 in this new ranking system --> modify this too
        if logs.get(self.monitor) is not None:
            curloss = logs.get(self.monitor)
            if curloss <= self.best_loss and not self.stop_me:
                # store path to the best model
                self.best_loss = curloss
                self.best_model = save_file_name
                logger.debug('best model at epoch ' + str(epoch) + ' ' + self.best_model)

            if self.save_training_log:
                f2 = open(self.save_output_name + '.log', 'a+')
                f2.write('epoch %03d - loss %s\n' % (epoch + 1, str(logs)))
                f2.close()
        return

    def on_batch_begin(self, batch, logs={}):
        return

    def on_batch_end(self, batch, logs={}):
        return

    def get_best_model(self):
        # return the past to the best model (lowest loss ever)
        return self.best_model



from skimage.filters import threshold_sauvola
from skimage.morphology import remove_small_objects
import matplotlib.pyplot as plt
from epyseg.img import Img
import numpy as np
from epyseg.postprocess.superpixel_methods import get_optimized_mask2


def sauvola(img, window_size=25, min_threshold=0.02):
    k = 0.25
    r = 0.5

    t = threshold_sauvola(img, window_size=window_size, k=k, r=r)
    if min_threshold is not None:
        t[t <= min_threshold] = min_threshold
    return t


def segment_cells(image, __DEBUG=False, __VISUAL_DEBUG=False, stop_at_threshold_step=False,
                  min_unconnected_object_size=None, min_threshold=None, window_size=25, real_avg_mode=False):
    original = image.copy()

    t = sauvola(image, min_threshold=min_threshold, window_size=window_size)

    if __VISUAL_DEBUG:
        plt.imshow(t)
        plt.show()

    image[image >= t] = 1
    image[image < t] = 0

    if __VISUAL_DEBUG and min_unconnected_object_size == 12:
        plt.imshow(image)
        plt.title('before')
        plt.show()

    if min_unconnected_object_size is not None and min_unconnected_object_size >= 1:
        # image = remove_small_objects(image.astype(bool), min_size=min_unconnected_object_size, connectivity=2, in_place=True).astype(np.uint8)
        # in the future I need to shift to that line (maybe I can add out in newer version but then it will not be compatible with previous --> keep it like that for now
        image = remove_small_objects(image.astype(bool), min_size=min_unconnected_object_size, connectivity=2).astype(np.uint8)
        if __VISUAL_DEBUG and min_unconnected_object_size == 12:
            plt.imshow(image)
            plt.title('after')
            plt.show()

    if stop_at_threshold_step:
        image = image * 255
        return image

    return get_optimized_mask2(original, sauvola_mask=None,
                               score_before_adding=True)# --> True


if __name__ == '__main__':
    from timeit import default_timer as timer

    # image = Img('/D/final_folder_scoring/predict_hybrid/mini_test.tif')
    # image = Img('/D/final_folder_scoring/predict_hybrid/AVG_StackFocused_Endocad-GFP(6-12-13)#19_000.tif')
    # image = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/122.tif')[...,0]
    # image = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/image_plant_best-zoomed.tif')[...,0]
    image = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/5.tif')[..., 0]
    # image = Img('/D/final_folder_scoring/predict_hybrid/tmp11.png')
    # image = Img('/D/final_folder_scoring/predict_hybrid/11-1_nuclei_1.tif')
    start = timer()

    # final_mask = segment_cells(image, __DEBUG=False, __VISUAL_DEBUG=False, stop_at_threshold_step=False)
    final_mask = segment_cells(image, __DEBUG=False, __VISUAL_DEBUG=False, stop_at_threshold_step=False, min_unconnected_object_size=12)

    duration = timer() - start
    print(duration)

    plt.imshow(final_mask)
    plt.show()
    Img(final_mask, dimensions='hw').save('/home/aigouy/Bureau/trash/test_new_seeds_seg_stuff/final_rewatershed.tif')



import logging

class TA_logger(object):

    # DEBUG < INFO < WARNING < ERROR < CRITICAL
    default_format = '%(levelname)s - %(asctime)s - %(filename)s - %(funcName)s - line %(lineno)d - %(message)s\n'
    master_logger_name = 'master'
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL
    DEFAULT = INFO # DEBUG # INFO

    loggers = {}

    def __new__(cls, name=master_logger_name, logging_level=DEFAULT, format=default_format, handler=None):
        if name is not None:
            if name in cls.loggers:
                return cls.loggers.get(name)

        logger = logging.getLogger(name)

        if handler is None:
            # create a formatter
            formatter = logging.Formatter(format)
            # create handler
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)

        logger.addHandler(handler)

        # set level to logging_level
        cls.DEFAULT = logging_level
        logger.setLevel(logging_level)
        cls.loggers[name] = logger

        return logger

    @staticmethod
    def setHandler(handler, name=master_logger_name):
        # easy way to redirect all logs to the same logger
        logger = logging.getLogger(name)
        try:
            for hndlr in logger.handlers:
                logger.removeHandler(hndlr)
        except:
            pass
        logger.addHandler(handler)
        logger.setLevel(TA_logger.DEFAULT)

if __name__ == '__main__':
    logger = TA_logger()
    logger.debug("test")
    logger.info("test")
    logger.warning("test")
    logger.error("test")
    logger.critical("test")

    formatter = logging.Formatter(TA_logger.default_format)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    TA_logger.setHandler(handler)
    logger.debug("test")
    logger.info("test")
    logger.warning("test")
    logger.error("test")
    logger.critical("test")

